# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import sys
import time
from collections import defaultdict
from typing import Any, Dict, List, Optional, Set, Type

from .... import oscar as mo
from ....core import ChunkGraph, OperandType, enter_mode, ExecutionError
from ....core.context import get_context, set_context
from ....core.operand import (
    Fetch,
    FetchShuffle,
    execute,
)
from ....metrics import Metrics
from ....optimization.physical import optimize
from ....typing import BandType, ChunkType
from ....utils import get_chunk_key_to_data_keys
from ...context import ThreadedServiceContext
from ...meta.api import MetaAPI, WorkerMetaAPI
from ...session import SessionAPI
from ...storage import StorageAPI
from ...task import TaskAPI, task_options
from ..core import Subtask, SubtaskStatus, SubtaskResult
from ..utils import iter_input_data_keys, iter_output_data, get_mapper_data_keys

logger = logging.getLogger(__name__)


class DataStore(dict):
    def __getattr__(self, attr):
        ctx = get_context()
        return getattr(ctx, attr)


BASIC_META_FIELDS = ["memory_size", "store_size", "bands", "object_ref"]


class SubtaskProcessor:
    _chunk_graph: ChunkGraph
    _chunk_key_to_data_keys: Dict[str, List[str]]

    def __init__(
        self,
        subtask: Subtask,
        session_api: SessionAPI,
        storage_api: StorageAPI,
        meta_api: MetaAPI,
        worker_meta_api: WorkerMetaAPI,
        band: BandType,
        supervisor_address: str,
        engines: List[str] = None,
    ):
        self.subtask = subtask
        self._session_id = self.subtask.session_id
        self._chunk_graph = subtask.chunk_graph
        self._actual_chunk_count = len(
            [
                chunk
                for chunk in subtask.chunk_graph
                if not isinstance(chunk.op, (Fetch, FetchShuffle))
            ]
        )
        self._band = band
        self._supervisor_address = supervisor_address
        self._engines = engines if engines is not None else task_options.runtime_engines

        # result
        self.result = SubtaskResult(
            subtask_id=subtask.subtask_id,
            session_id=subtask.session_id,
            task_id=subtask.task_id,
            stage_id=subtask.stage_id,
            status=SubtaskStatus.pending,
            bands=[self._band],
            progress=0.0,
            execution_start_time=time.time(),
        )
        self.is_done = asyncio.Event()

        # status and intermediate states
        # operand progress, from op key to progress
        self._op_progress: Dict[str, float] = defaultdict(lambda: 0.0)
        # temp data store that holds chunk data during computation
        self._datastore = DataStore()
        # chunk key to real data keys
        self._chunk_key_to_data_keys = dict()

        # other service APIs
        self._session_api = session_api
        self._storage_api = storage_api
        self._meta_api = meta_api
        self._worker_meta_api = worker_meta_api

        # add metrics
        self._subtask_execution_time = Metrics.gauge(
            "mars.subtask_execution_time_secs",
            "Time consuming in seconds to execute a subtask",
            ("session_id", "subtask_id"),
        )

    @property
    def status(self):
        return self.result.status

    @property
    def subtask_id(self):
        return self.subtask.subtask_id

    async def _load_input_data(self):
        keys, gets, accept_nones = [], [], []
        for key, is_shuffle in iter_input_data_keys(
            self.subtask, self._chunk_graph, self._chunk_key_to_data_keys
        ):
            keys.append(key)
            accept_nones.append(not is_shuffle)
            gets_params = {"error": "ignore"} if is_shuffle else {}
            gets.append(self._storage_api.get.delay(key, **gets_params))
        if keys:
            logger.debug(
                "Start getting input data, keys: %s, subtask id: %s",
                keys,
                self.subtask.subtask_id,
            )
            inputs = await self._storage_api.get.batch(*gets)
            self._datastore.update(
                {
                    key: get
                    for key, get, accept_none in zip(keys, inputs, accept_nones)
                    if accept_none or get is not None
                }
            )
            logger.debug(
                "Finish getting input data keys: %s, subtask id: %s",
                keys,
                self.subtask.subtask_id,
            )
        return keys

    @staticmethod
    async def notify_task_manager_result(
        supervisor_address: str, result: SubtaskResult
    ):
        task_api = await TaskAPI.create(result.session_id, supervisor_address)
        # notify task service
        await task_api.set_subtask_result(result)

    def _init_ref_counts(self) -> Dict[str, int]:
        chunk_graph = self._chunk_graph
        ref_counts = defaultdict(lambda: 0)
        # set 1 for result chunks
        for result_chunk in chunk_graph.result_chunks:
            ref_counts[result_chunk.key] += 1
        # iter graph to set ref counts
        for chunk in chunk_graph:
            ref_counts[chunk.key] += chunk_graph.count_successors(chunk)
        return ref_counts

    async def _async_execute_operand(self, ctx: Dict[str, Any], op: OperandType):
        if not isinstance(op, (Fetch, FetchShuffle)):
            self._op_progress[op.key] = 0.0
        get_context().set_running_operand_key(self._session_id, op.key)
        return asyncio.to_thread(self._execute_operand, ctx, op)

    def set_op_progress(self, op_key: str, progress: float):
        if op_key in self._op_progress:  # pragma: no branch
            self._op_progress[op_key] = progress

    @enter_mode(build=False, kernel=True)
    def _execute_operand(
        self, ctx: Dict[str, Any], op: OperandType
    ):  # noqa: R0201  # pylint: disable=no-self-use
        try:
            return execute(ctx, op)
        except BaseException as ex:
            # wrap exception in execution to avoid side effects
            raise ExecutionError(ex).with_traceback(ex.__traceback__) from None

    async def _execute_graph(self, chunk_graph: ChunkGraph):
        loop = asyncio.get_running_loop()
        ref_counts = self._init_ref_counts()

        # from data_key to results
        for chunk in chunk_graph.topological_iter():
            if chunk.key not in self._datastore:
                # since `op.execute` may be a time-consuming operation,
                # we make it run in a thread pool to not block current thread.
                logger.debug(
                    "Start executing operand: %s, chunk: %s, subtask id: %s",
                    chunk.op,
                    chunk,
                    self.subtask.subtask_id,
                )
                future = asyncio.create_task(
                    await self._async_execute_operand(self._datastore, chunk.op)
                )
                to_wait = loop.create_future()

                def cb(fut):
                    if not to_wait.done():
                        if fut.exception():
                            to_wait.set_exception(fut.exception())
                        else:
                            to_wait.set_result(fut.result())

                future.add_done_callback(cb)

                try:
                    await to_wait
                    logger.debug(
                        "Finish executing operand: %s, chunk: %s, subtask id: %s",
                        chunk.op,
                        chunk,
                        self.subtask.subtask_id,
                    )
                except asyncio.CancelledError:
                    logger.debug(
                        "Receive cancel instruction for operand: %s,"
                        "chunk: %s, subtask id: %s",
                        chunk.op,
                        chunk,
                        self.subtask.subtask_id,
                    )
                    # wait for this computation to finish
                    await future
                    # if cancelled, stop next computation
                    logger.debug(
                        "Cancelled operand: %s, chunk: %s, subtask id: %s",
                        chunk.op,
                        chunk,
                        self.subtask.subtask_id,
                    )
                    self.result.status = SubtaskStatus.cancelled
                    raise

            self.set_op_progress(chunk.op.key, 1.0)

            for inp in chunk_graph.iter_predecessors(chunk):
                ref_counts[inp.key] -= 1
                if ref_counts[inp.key] == 0:
                    # ref count reaches 0, remove it
                    for key in self._chunk_key_to_data_keys[inp.key]:
                        if key in self._datastore:
                            del self._datastore[key]

    async def _unpin_data(self, data_keys):
        # unpin input keys
        unpins = []
        shuffle_unpins = []
        for key in data_keys:
            if isinstance(key, tuple):
                # a tuple key means it's a shuffle key,
                # some shuffle data is None and not stored in storage
                shuffle_unpins.append(
                    self._storage_api.unpin.delay(key, error="ignore")
                )
            else:
                unpins.append(self._storage_api.unpin.delay(key))
        if unpins:
            await self._storage_api.unpin.batch(*unpins)
        if shuffle_unpins:
            # TODO(hks): The batch method doesn't accept different error arguments,
            #  combine them when it can.
            await self._storage_api.unpin.batch(*shuffle_unpins)

    async def _store_data(self, chunk_graph: ChunkGraph):
        # store data into storage
        data_key_to_puts = {}
        for key, data, _ in iter_output_data(chunk_graph, self._datastore):
            put = self._storage_api.put.delay(key, data)
            data_key_to_puts[key] = put

        stored_keys = list(data_key_to_puts.keys())
        puts = data_key_to_puts.values()
        logger.debug(
            "Start putting data keys: %s, subtask id: %s",
            stored_keys,
            self.subtask.subtask_id,
        )
        data_key_to_store_size = dict()
        data_key_to_memory_size = dict()
        data_key_to_object_id = dict()
        if puts:
            put_infos = asyncio.create_task(self._storage_api.put.batch(*puts))
            try:
                store_infos = await put_infos
                for store_key, store_info in zip(stored_keys, store_infos):
                    data_key_to_store_size[store_key] = store_info.store_size
                    data_key_to_memory_size[store_key] = store_info.memory_size
                    data_key_to_object_id[store_key] = store_info.object_id
                logger.debug(
                    "Finish putting data keys: %s, subtask id: %s",
                    stored_keys,
                    self.subtask.subtask_id,
                )
            except asyncio.CancelledError:
                logger.debug(
                    "Cancelling put data keys: %s, subtask id: %s",
                    stored_keys,
                    self.subtask.subtask_id,
                )
                put_infos.cancel()

                logger.debug(
                    "Cancelled put data keys: %s, subtask id: %s",
                    stored_keys,
                    self.subtask.subtask_id,
                )
                self.result.status = SubtaskStatus.cancelled
                raise

        # clear data
        self._datastore = dict()
        return (
            stored_keys,
            data_key_to_store_size,
            data_key_to_memory_size,
            data_key_to_object_id,
        )

    async def _store_meta(
        self,
        chunk_graph: ChunkGraph,
        data_key_to_store_size: Dict,
        data_key_to_memory_size: Dict,
        data_key_to_object_id: Dict,
        update_meta_chunks: Set[ChunkType],
    ):
        # store meta
        set_chunk_metas = []
        set_worker_chunk_metas = []
        result_data_size = 0
        set_meta_keys = []
        for result_chunk in chunk_graph.result_chunks:
            chunk_key = result_chunk.key
            set_meta_keys.append(chunk_key)
            if chunk_key in data_key_to_store_size:
                # normal chunk
                store_size = data_key_to_store_size[chunk_key]
                memory_size = data_key_to_memory_size[chunk_key]
                result_data_size += memory_size
                object_ref = data_key_to_object_id[chunk_key]
            else:
                # mapper chunk
                mapper_keys = get_mapper_data_keys(chunk_key, data_key_to_store_size)
                store_size = sum(data_key_to_store_size[k] for k in mapper_keys)
                memory_size = sum(data_key_to_memory_size[k] for k in mapper_keys)
                object_ref = [data_key_to_object_id[k] for k in mapper_keys]
            # for worker, if chunk in update_meta_chunks
            # save meta including dtypes_value etc, otherwise,
            # save basic meta only
            if result_chunk in update_meta_chunks:
                set_worker_chunk_metas.append(
                    self._worker_meta_api.set_chunk_meta.delay(
                        result_chunk,
                        memory_size=memory_size,
                        store_size=store_size,
                        bands=[self._band],
                        chunk_key=chunk_key,
                        exclude_fields=["object_ref"],
                    )
                )
            # for supervisor, only save basic meta that is small like memory_size etc
            set_chunk_metas.append(
                self._meta_api.set_chunk_meta.delay(
                    result_chunk,
                    memory_size=memory_size,
                    store_size=store_size,
                    bands=[self._band],
                    chunk_key=chunk_key,
                    object_ref=object_ref,
                    fields=BASIC_META_FIELDS,
                )
            )
        logger.debug(
            "Start storing chunk metas for data keys: %s, subtask id: %s",
            set_meta_keys,
            self.subtask.subtask_id,
        )
        if set_chunk_metas:
            f = asyncio.get_running_loop().create_future()

            async def set_chunks_meta():
                coros = []
                if set_worker_chunk_metas:
                    coros.append(
                        self._worker_meta_api.set_chunk_meta.batch(
                            *set_worker_chunk_metas
                        )
                    )
                coros.append(self._meta_api.set_chunk_meta.batch(*set_chunk_metas))
                await asyncio.gather(*coros)
                logger.debug(
                    "Finish store chunk metas for data keys: %s, subtask id: %s",
                    set_meta_keys,
                    self.subtask.subtask_id,
                )
                f.set_result(None)

            try:
                # Since we don't delete chunk data on this worker,
                # we need to ensure chunk meta are recorded
                # in meta service, so that `processor.decref_stage`
                # can delete the chunk data finally.
                await asyncio.shield(set_chunks_meta())
            except asyncio.CancelledError:  # pragma: no cover
                await f
                raise
        # set result data size
        self.result.data_size = result_data_size

    async def done(self):
        if self.result.status == SubtaskStatus.running:
            self.result.status = SubtaskStatus.succeeded
            # Only update end time when subtask succeeded
            self.result.execution_end_time = time.time()
        self.result.progress = 1.0
        self.is_done.set()

    async def run(self):
        self.result.status = SubtaskStatus.running
        input_keys = None
        unpinned = False
        try:
            raw_result_chunks = list(self._chunk_graph.result_chunks)
            chunk_graph = optimize(self._chunk_graph, self._engines)
            self._chunk_key_to_data_keys = get_chunk_key_to_data_keys(chunk_graph)
            report_progress = asyncio.create_task(self.report_progress_periodically())

            result_chunk_to_optimized = {
                c: o for c, o in zip(raw_result_chunks, chunk_graph.result_chunks)
            }
            raw_update_meta_chunks = self.subtask.update_meta_chunks
            if raw_update_meta_chunks is None:
                raw_update_meta_chunks = raw_result_chunks
            update_meta_chunks = {
                result_chunk_to_optimized[c] for c in raw_update_meta_chunks
            }

            # load inputs data
            input_keys = await self._load_input_data()
            try:
                # execute chunk graph
                await self._execute_graph(chunk_graph)
            finally:
                # unpin inputs data
                unpinned = True
                await self._unpin_data(input_keys)
            # store results data
            (
                stored_keys,
                store_sizes,
                memory_sizes,
                data_key_to_object_id,
            ) = await self._store_data(chunk_graph)
            # store meta
            await self._store_meta(
                chunk_graph,
                store_sizes,
                memory_sizes,
                data_key_to_object_id,
                update_meta_chunks,
            )
        except asyncio.CancelledError:
            self.result.status = SubtaskStatus.cancelled
            self.result.progress = 1.0
            raise
        except BaseException as ex:  # noqa: E722  # nosec  # pylint: disable=bare-except
            self.result.status = SubtaskStatus.errored
            self.result.progress = 1.0
            if isinstance(ex, ExecutionError):
                self.result.error = ex.nested_error
                self.result.traceback = ex.nested_error.__traceback__
            else:  # pragma: no cover
                _, self.result.error, self.result.traceback = sys.exc_info()
            await self.done()
            raise
        finally:
            if input_keys is not None and not unpinned:
                await self._unpin_data(input_keys)

        await self.done()
        if self.result.status == SubtaskStatus.succeeded:
            cost_time_secs = (
                self.result.execution_end_time - self.result.execution_start_time
            )
            logger.info(
                "Time consuming to execute a subtask is %ss with session_id %s, subtask_id %s",
                cost_time_secs,
                self._session_id,
                self.subtask.subtask_id,
            )
            self._subtask_execution_time.record(
                cost_time_secs,
                {"session_id": self._session_id, "subtask_id": self.subtask.subtask_id},
            )
        report_progress.cancel()
        try:
            await report_progress
        except asyncio.CancelledError:
            pass
        return self.result

    async def report_progress_periodically(self, interval=0.5, eps=0.001):
        last_progress = self.result.progress
        while not self.result.status.is_done:
            size = self._actual_chunk_count
            progress = sum(self._op_progress.values()) / size
            assert progress <= 1
            self.result.progress = progress
            if abs(last_progress - progress) >= eps:
                # report progress
                if not self.result.status.is_done:
                    fut = self.notify_task_manager_result(
                        self._supervisor_address, self.result
                    )
                    if fut:
                        await fut
            await asyncio.sleep(interval)
            last_progress = progress


class SubtaskProcessorActor(mo.Actor):
    _session_api: Optional[SessionAPI]
    _storage_api: Optional[StorageAPI]
    _meta_api: Optional[MetaAPI]
    _worker_meta_api: Optional[WorkerMetaAPI]
    _processor: Optional[SubtaskProcessor]
    _last_processor: Optional[SubtaskProcessor]
    _running_aio_task: Optional[asyncio.Task]

    def __init__(
        self,
        session_id: str,
        band: BandType,
        supervisor_address: str,
        worker_address: str,
        subtask_processor_cls: Type[SubtaskProcessor],
    ):
        self._session_id = session_id
        self._band = band
        self._supervisor_address = supervisor_address
        self._worker_address = worker_address
        self._subtask_processor_cls = subtask_processor_cls

        # current processor
        self._processor = None
        self._last_processor = None
        self._running_aio_task = None

        self._session_api = None
        self._storage_api = None
        self._meta_api = None
        self._worker_meta_api = None

    @classmethod
    def gen_uid(cls, session_id: str):
        return f"{session_id}_subtask_processor"

    async def __post_create__(self):
        coros = [
            SessionAPI.create(self._supervisor_address),
            StorageAPI.create(self._session_id, self.address, self._band[1]),
            MetaAPI.create(self._session_id, self._supervisor_address),
            WorkerMetaAPI.create(self._session_id, self.address),
        ]
        coros = [asyncio.ensure_future(coro) for coro in coros]
        await asyncio.gather(*coros)
        self._session_api, self._storage_api, self._meta_api, self._worker_meta_api = [
            coro.result() for coro in coros
        ]

    async def _init_context(self, session_id: str):
        loop = asyncio.get_running_loop()
        context = ThreadedServiceContext(
            session_id,
            self._supervisor_address,
            self._worker_address,
            self.address,
            loop,
            band=self._band,
        )
        await context.init()
        set_context(context)

    async def run(self, subtask: Subtask):
        logger.info(
            "Start to run subtask: %r on %s. chunk graph contains %s",
            subtask,
            self.address,
            [c for c in subtask.chunk_graph],
        )

        assert subtask.session_id == self._session_id

        # init context
        await self._init_context(self._session_id)
        processor = self._subtask_processor_cls(
            subtask,
            self._session_api,
            self._storage_api,
            self._meta_api,
            self._worker_meta_api,
            self._band,
            self._supervisor_address,
        )
        self._processor = self._last_processor = processor
        self._running_aio_task = asyncio.create_task(processor.run())
        try:
            result = yield self._running_aio_task
            logger.info("Finished subtask: %s", subtask.subtask_id)
            raise mo.Return(result)
        finally:
            self._processor = self._running_aio_task = None

    async def wait(self):
        return self._processor.is_done.wait()

    async def result(self):
        return self._last_processor.result

    async def cancel(self):
        logger.info("Cancelling subtask: %s", self._processor.subtask_id)

        aio_task = self._running_aio_task
        aio_task.cancel()

        async def waiter():
            try:
                await aio_task
            except asyncio.CancelledError:
                pass

        # return asyncio task to not block current actor
        return waiter()

    def get_running_subtask_id(self):
        return self._processor.subtask_id

    def set_running_op_progress(self, op_key: str, progress: float):
        self._processor.set_op_progress(op_key, progress)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable
from functools import reduce
from operator import and_

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, AnyField
from ...lib.sparse import SparseNDArray
from ...lib.sparse.core import naked, cps, sps
from .core import TensorHasInput
from .array import tensor


class DenseToSparse(TensorHasInput):
    _op_type_ = OperandDef.DENSE_TO_SPARSE

    _input = KeyField("input")
    _missing = AnyField("missing")

    def __init__(self, missing=None, **kw):
        super().__init__(sparse=True, _missing=missing, **kw)

    @property
    def missing(self):
        return self._missing

    @staticmethod
    def _get_mask(data, missing):
        if isinstance(missing, Iterable):
            return reduce(and_, (DenseToSparse._get_mask(data, m) for m in missing))
        elif pd.isna(missing):
            return ~pd.isna(data)
        else:
            return data != missing

    @classmethod
    def execute(cls, ctx, op):
        out = op.outputs[0]
        in_data = naked(ctx[op.inputs[0].key])
        missing = op.missing
        shape = in_data.shape if any(np.isnan(s) for s in out.shape) else out.shape

        xps = cps if op.gpu else sps
        if missing is None:
            ctx[out.key] = SparseNDArray(xps.csr_matrix(in_data), shape=shape)
        else:
            mask = cls._get_mask(in_data, missing)
            spmatrix = xps.csr_matrix((in_data[mask], mask.nonzero()), shape=shape)
            ctx[out.key] = SparseNDArray(spmatrix)


def fromdense(a, missing=None):
    a = tensor(a)
    if a.issparse():
        return a

    op = DenseToSparse(dtype=a.dtype, gpu=a.op.gpu, missing=missing)
    return op(a)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Tuple, Type, TypeVar, Union

from .... import oscar as mo
from ....lib.aio import alru_cache
from ...subtask import Subtask
from ..core import SubtaskScheduleSummary
from .core import AbstractSchedulingAPI

APIType = TypeVar("APIType", bound="SchedulingAPI")


class SchedulingAPI(AbstractSchedulingAPI):
    def __init__(
        self, session_id: str, address: str, manager_ref=None, queueing_ref=None
    ):
        self._session_id = session_id
        self._address = address

        self._manager_ref = manager_ref
        self._queueing_ref = queueing_ref

    @classmethod
    @alru_cache
    async def create(cls: Type[APIType], session_id: str, address: str) -> APIType:
        from ..supervisor.manager import SubtaskManagerActor

        manager_ref = await mo.actor_ref(
            SubtaskManagerActor.gen_uid(session_id), address=address
        )
        from ..supervisor.queueing import SubtaskQueueingActor

        queueing_ref = await mo.actor_ref(
            SubtaskQueueingActor.gen_uid(session_id), address=address
        )

        scheduling_api = SchedulingAPI(session_id, address, manager_ref, queueing_ref)
        return scheduling_api

    async def get_subtask_schedule_summaries(
        self, task_id: Optional[str] = None
    ) -> List[SubtaskScheduleSummary]:
        return await self._manager_ref.get_schedule_summaries(task_id)

    async def add_subtasks(
        self, subtasks: List[Subtask], priorities: Optional[List[Tuple]] = None
    ):
        """
        Submit subtasks into scheduling service

        Parameters
        ----------
        subtasks
            list of subtasks to be submitted to service
        priorities
            list of priorities of subtasks
        """
        if priorities is None:
            priorities = [subtask.priority or tuple() for subtask in subtasks]
        await self._manager_ref.add_subtasks(subtasks, priorities)

    @mo.extensible
    async def update_subtask_priority(self, subtask_id: str, priority: Tuple):
        """
        Update priorities of subtasks

        Parameters
        ----------
        subtask_id
            id of subtask to update priority
        priority
            list of priority of subtasks
        """
        raise NotImplementedError

    @update_subtask_priority.batch
    async def update_subtask_priority(self, args_list, kwargs_list):
        await self._queueing_ref.update_subtask_priority.batch(
            *(
                self._queueing_ref.update_subtask_priority.delay(*args, **kwargs)
                for args, kwargs in zip(args_list, kwargs_list)
            )
        )

    async def cancel_subtasks(
        self, subtask_ids: List[str], kill_timeout: Union[float, int] = None
    ):
        """
        Cancel pending and running subtasks.

        Parameters
        ----------
        subtask_ids
            ids of subtasks to cancel
        kill_timeout
            timeout seconds to kill actor process forcibly
        """
        await self._manager_ref.cancel_subtasks(subtask_ids, kill_timeout=kill_timeout)

    async def finish_subtasks(
        self,
        subtask_ids: List[str],
        bands: List[Tuple] = None,
        schedule_next: bool = True,
    ):
        """
        Mark subtasks as finished, letting scheduling service to schedule
        next tasks in the ready queue

        Parameters
        ----------
        subtask_ids
            ids of subtasks to mark as finished
        bands
            bands of subtasks to mark as finished
        schedule_next
            whether to schedule succeeding subtasks
        """
        await self._manager_ref.finish_subtasks(subtask_ids, bands, schedule_next)


class MockSchedulingAPI(SchedulingAPI):
    @classmethod
    async def create(cls: Type[APIType], session_id: str, address: str) -> APIType:
        from ..supervisor import GlobalResourceManagerActor, AutoscalerActor

        await mo.create_actor(
            GlobalResourceManagerActor,
            uid=GlobalResourceManagerActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            AutoscalerActor, {}, uid=AutoscalerActor.default_uid(), address=address
        )

        from .... import resource as mars_resource
        from ..worker import (
            SubtaskExecutionActor,
            WorkerSlotManagerActor,
            WorkerQuotaManagerActor,
        )

        await mo.create_actor(
            SubtaskExecutionActor,
            subtask_max_retries=0,
            uid=SubtaskExecutionActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            WorkerSlotManagerActor,
            uid=WorkerSlotManagerActor.default_uid(),
            address=address,
        )
        await mo.create_actor(
            WorkerQuotaManagerActor,
            {"quota_size": mars_resource.virtual_memory().total},
            uid=WorkerQuotaManagerActor.default_uid(),
            address=address,
        )

        from ..supervisor import SchedulingSupervisorService

        service = SchedulingSupervisorService({}, address)
        await service.create_session(session_id)
        return await super().create(session_id, address)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

from contextlib import contextmanager
from enum import Enum
from queue import PriorityQueue
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple

from .backends.console import console_metric
from .backends.prometheus import prometheus_metric
from .backends.ray import ray_metric

logger = logging.getLogger(__name__)

_init = False
_metric_backend = "console"
_backends_cls = {
    "console": console_metric,
    "prometheus": prometheus_metric,
    "ray": ray_metric,
}


def init_metrics(backend="console", config: Dict[str, Any] = None):
    global _init
    if _init is True:
        return

    backend = backend or "console"
    if backend not in _backends_cls:
        raise NotImplementedError(f"Do not support metric backend {backend}")
    global _metric_backend
    _metric_backend = backend
    if _metric_backend == "prometheus":
        try:
            from prometheus_client import start_http_server
            from ..utils import get_next_port

            port = config.get("port", 0) if config else 0
            port = port or get_next_port()
            start_http_server(port)
            logger.warning(
                "Finished startup prometheus http server and port is %d", port
            )
        except ImportError:
            logger.warning(
                "Failed to start prometheus http server because there is no prometheus_client"
            )
    _init = True
    logger.info("Finished initialize the metrics with backend %s", _metric_backend)


def shutdown_metrics():
    global _metric_backend
    _metric_backend = "console"
    global _init
    _init = False


class Metrics:
    """
    A factory to generate different types of metrics.

    Note:
        Counter, Meter and Histogram are not thread safe.

    Examples
    --------
    >>> c1 = counter('counter1', 'A counter')
    >>> c1.record(1)

    >>> c2 = counter('counter2', 'A counter', ('service', 'tenant'))
    >>> c2.record(1, {'service': 'mars', 'tenant': 'test'})

    >>> g1 = gauge('gauge1')
    >>> g1.record(1)

    >>> m1 = meter('meter1')
    >>> m1.record(1)

    >>> h1 = histogram('histogram1')
    >>> h1.record(1)
    """

    @staticmethod
    def counter(name, description: str = "", tag_keys: Optional[Tuple[str]] = None):
        return _backends_cls[_metric_backend].Counter(name, description, tag_keys)

    @staticmethod
    def gauge(name, description: str = "", tag_keys: Optional[Tuple[str]] = None):
        return _backends_cls[_metric_backend].Gauge(name, description, tag_keys)

    @staticmethod
    def meter(name, description: str = "", tag_keys: Optional[Tuple[str]] = None):
        return _backends_cls[_metric_backend].Meter(name, description, tag_keys)

    @staticmethod
    def histogram(name, description: str = "", tag_keys: Optional[Tuple[str]] = None):
        return _backends_cls[_metric_backend].Histogram(name, description, tag_keys)


class Percentile:
    class PercentileType(Enum):
        P99 = 1
        P95 = 2
        P90 = 3

    def __init__(self, capacity: int, window: int, callback: Callable[[float], None]):
        self._capacity = capacity
        self._window = window
        self._callback = callback
        self._min_heap = PriorityQueue()
        self._cur_num = 0

        if capacity <= 0 or window <= 0:
            raise ValueError(
                f"capacity or window expect to get a positive integer,"
                f"but capacity got: {capacity} and window got: {window}"
            )

    def record_data(self, value):
        store_value = -1 * value
        if self._min_heap.qsize() < self._capacity:
            self._min_heap.put(store_value)
        else:
            top_value = self._min_heap.get_nowait()
            store_value = store_value if top_value < store_value else top_value
            self._min_heap.put(store_value)

        self._cur_num += 1
        if self._cur_num % self._window == 0:
            self._callback(-1 * self._min_heap.get_nowait())
            self._cur_num = 0
            self._min_heap = PriorityQueue()

    @classmethod
    def build_p99(cls, callback: Callable[[float], None], window: int):
        return cls(int(window * 0.01), window, callback)

    @classmethod
    def build_p95(cls, callback: Callable[[float], None], window: int):
        return cls(int(window * 0.05), window, callback)

    @classmethod
    def build_p90(cls, callback: Callable[[float], None], window: int):
        return cls(int(window * 0.1), window, callback)


_percentile_builder = {
    Percentile.PercentileType.P99: Percentile.build_p99,
    Percentile.PercentileType.P95: Percentile.build_p95,
    Percentile.PercentileType.P90: Percentile.build_p90,
}


class PercentileArg(NamedTuple):
    percentile_type: Percentile.PercentileType
    callback: Callable[[float], None]
    window: int


@contextmanager
def record_time_cost_percentile(percentile_args: List[PercentileArg]):
    percentile_list = [
        _percentile_builder[percentile_type](callback, window)
        for percentile_type, callback, window in percentile_args
    ]
    st_time = time.time()

    yield

    cost_time = time.time() - st_time
    for percentile in percentile_list:
        percentile.record_data(cost_time)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import itertools
import operator
import numpy as np

from ..datasource.empty import empty
from ..datasource.array import array

# Internal functions to eliminate the overhead of repeated dispatch in one of
# the two possible paths inside mt.block.
# Use getattr to protect against __array_function__ being disabled.
_size = getattr(np.size, "__wrapped__", np.size)
_ndim = getattr(np.ndim, "__wrapped__", np.ndim)


def _block_format_index(index):
    """
    Convert a list of indices ``[0, 1, 2]`` into ``"arrays[0][1][2]"``.
    """
    idx_str = "".join("[{}]".format(i) for i in index if i is not None)
    return "arrays" + idx_str


def _block_check_depths_match(arrays, parent_index=[]):
    """
    Recursive function checking that the depths of nested lists in `arrays`
    all match. Mismatch raises a ValueError as described in the block
    docstring below.

    The entire index (rather than just the depth) needs to be calculated
    for each innermost list, in case an error needs to be raised, so that
    the index of the offending list can be printed as part of the error.

    Parameters
    ----------
    arrays : nested list of arrays
        The arrays to check
    parent_index : list of int
        The full index of `arrays` within the nested lists passed to
        `_block_check_depths_match` at the top of the recursion.

    Returns
    -------
    first_index : list of int
        The full index of an element from the bottom of the nesting in
        `arrays`. If any element at the bottom is an empty list, this will
        refer to it, and the last index along the empty axis will be None.
    max_arr_ndim : int
        The maximum of the ndims of the arrays nested in `arrays`.
    final_size: int
        The number of elements in the final array. This is used the motivate
        the choice of algorithm used using benchmarking wisdom.

    """
    if type(arrays) is tuple:
        # not strictly necessary, but saves us from:
        #  - more than one way to do things - no point treating tuples like
        #    lists
        #  - horribly confusing behaviour that results when tuples are
        #    treated like ndarray
        raise TypeError(
            "{} is a tuple. "
            "Only lists can be used to arrange blocks, and mt.block does "
            "not allow implicit conversion from tuple to ndarray.".format(
                _block_format_index(parent_index)
            )
        )
    elif type(arrays) is list and len(arrays) > 0:
        idxs_ndims = (
            _block_check_depths_match(arr, parent_index + [i])
            for i, arr in enumerate(arrays)
        )

        first_index, max_arr_ndim, final_size = next(idxs_ndims)
        for index, ndim, size in idxs_ndims:
            final_size += size
            if ndim > max_arr_ndim:
                max_arr_ndim = ndim
            if len(index) != len(first_index):
                raise ValueError(
                    "List depths are mismatched. First element was at depth "
                    "{}, but there is an element at depth {} ({})".format(
                        len(first_index), len(index), _block_format_index(index)
                    )
                )
            # propagate our flag that indicates an empty list at the bottom
            if index[-1] is None:
                first_index = index

        return first_index, max_arr_ndim, final_size
    elif type(arrays) is list and len(arrays) == 0:
        # We've 'bottomed out' on an empty list
        return parent_index + [None], 0, 0
    else:
        # We've 'bottomed out' - arrays is either a scalar or an array
        size = _size(arrays)
        return parent_index, _ndim(arrays), size


def _atleast_nd(a, ndim):
    # Ensures `a` has at least `ndim` dimensions by prepending
    # ones to `a.shape` as necessary
    return array(a, ndmin=ndim, copy=False)


def _accumulate(values):
    return list(itertools.accumulate(values))


def _concatenate_shapes(shapes, axis):
    """Given array shapes, return the resulting shape and slices prefixes.
    These help in nested concatenation.

    Returns
    -------
    shape: tuple of int
        This tuple satisfies:
        ```
        shape, _ = _concatenate_shapes([arr.shape for shape in arrs], axis)
        shape == concatenate(arrs, axis).shape
        ```
    slice_prefixes: tuple of (slice(start, end), )
        For a list of arrays being concatenated, this returns the slice
        in the larger array at axis that needs to be sliced into.
        For example, the following holds:
        ```
        ret = concatenate([a, b, c], axis)
        _, (sl_a, sl_b, sl_c) = concatenate_slices([a, b, c], axis)
        ret[(slice(None),) * axis + sl_a] == a
        ret[(slice(None),) * axis + sl_b] == b
        ret[(slice(None),) * axis + sl_c] == c
        ```
        These are called slice prefixes since they are used in the recursive
        blocking algorithm to compute the left-most slices during the
        recursion. Therefore, they must be prepended to rest of the slice
        that was computed deeper in the recursion.
        These are returned as tuples to ensure that they can quickly be added
        to existing slice tuple without creating a new tuple every time.
    """
    # Cache a result that will be reused.
    shape_at_axis = [shape[axis] for shape in shapes]

    # Take a shape, any shape
    first_shape = shapes[0]
    first_shape_pre = first_shape[:axis]
    first_shape_post = first_shape[axis + 1 :]

    if any(
        shape[:axis] != first_shape_pre or shape[axis + 1 :] != first_shape_post
        for shape in shapes
    ):
        raise ValueError("Mismatched array shapes in block along axis {}.".format(axis))

    shape = first_shape_pre + (sum(shape_at_axis),) + first_shape[axis + 1 :]

    offsets_at_axis = _accumulate(shape_at_axis)
    slice_prefixes = [
        (slice(start, end),)
        for start, end in zip([0] + offsets_at_axis, offsets_at_axis)
    ]
    return shape, slice_prefixes


def _block_info_recursion(arrays, max_depth, result_ndim, depth=0):
    """
    Returns the shape of the final array, along with a list
    of slices and a list of arrays that can be used for assignment inside the
    new array

    Parameters
    ----------
    arrays : nested list of arrays
        The arrays to check
    max_depth : list of int
        The number of nested lists
    result_ndim: int
        The number of dimensions in thefinal array.

    Returns
    -------
    shape : tuple of int
        The shape that the final array will take on.
    slices: list of tuple of slices
        The slices into the full array required for assignment. These are
        required to be prepended with ``(Ellipsis, )`` to obtain to correct
        final index.
    arrays: list of ndarray
        The data to assign to each slice of the full array

    """
    if depth < max_depth:
        shapes, slices, arrays = zip(
            *[
                _block_info_recursion(arr, max_depth, result_ndim, depth + 1)
                for arr in arrays
            ]
        )

        axis = result_ndim - max_depth + depth
        shape, slice_prefixes = _concatenate_shapes(shapes, axis)

        # Prepend the slice prefix and flatten the slices
        slices = [
            slice_prefix + the_slice
            for slice_prefix, inner_slices in zip(slice_prefixes, slices)
            for the_slice in inner_slices
        ]

        # Flatten the array list
        arrays = functools.reduce(operator.add, arrays)

        return shape, slices, arrays
    else:
        # We've 'bottomed out' - arrays is either a scalar or an array
        # type(arrays) is not list
        # Return the slice and the array inside a list to be consistent with
        # the recursive case.
        arr = _atleast_nd(arrays, result_ndim)
        return arr.shape, [()], [arr]


def _block(arrays, max_depth, result_ndim, depth=0):
    """
    Internal implementation of block based on repeated concatenation.
    `arrays` is the argument passed to
    block. `max_depth` is the depth of nested lists within `arrays` and
    `result_ndim` is the greatest of the dimensions of the arrays in
    `arrays` and the depth of the lists in `arrays` (see block docstring
    for details).
    """
    from ..merge.concatenate import concatenate

    if depth < max_depth:
        arrs = [_block(arr, max_depth, result_ndim, depth + 1) for arr in arrays]
        return concatenate(arrs, axis=-(max_depth - depth))
    else:
        # We've 'bottomed out' - arrays is either a scalar or an array
        # type(arrays) is not list
        return _atleast_nd(arrays, result_ndim)


def block(arrays):
    """
    Assemble an nd-array from nested lists of blocks.

    Blocks in the innermost lists are concatenated (see `concatenate`) along
    the last dimension (-1), then these are concatenated along the
    second-last dimension (-2), and so on until the outermost list is reached.

    Blocks can be of any dimension, but will not be broadcasted using the normal
    rules. Instead, leading axes of size 1 are inserted, to make ``block.ndim``
    the same for all blocks. This is primarily useful for working with scalars,
    and means that code like ``mt.block([v, 1])`` is valid, where
    ``v.ndim == 1``.

    When the nested list is two levels deep, this allows block matrices to be
    constructed from their components.

    .. versionadded:: 1.13.0

    Parameters
    ----------
    arrays : nested list of array_like or scalars (but not tuples)
        If passed a single ndarray or scalar (a nested list of depth 0), this
        is returned unmodified (and not copied).

        Elements shapes must match along the appropriate axes (without
        broadcasting), but leading 1s will be prepended to the shape as
        necessary to make the dimensions match.

    Returns
    -------
    block_array : Tensor
        The array assembled from the given blocks.

        The dimensionality of the output is equal to the greatest of:
        * the dimensionality of all the inputs
        * the depth to which the input list is nested

    Raises
    ------
    ValueError
        * If list depths are mismatched - for instance, ``[[a, b], c]`` is
          illegal, and should be spelt ``[[a, b], [c]]``
        * If lists are empty - for instance, ``[[a, b], []]``

    See Also
    --------
    concatenate : Join a sequence of arrays along an existing axis.
    stack : Join a sequence of arrays along a new axis.
    vstack : Stack arrays in sequence vertically (row wise).
    hstack : Stack arrays in sequence horizontally (column wise).
    dstack : Stack arrays in sequence depth wise (along third axis).
    column_stack : Stack 1-D arrays as columns into a 2-D array.
    vsplit : Split an array into multiple sub-arrays vertically (row-wise).

    Notes
    -----

    When called with only scalars, ``mt.block`` is equivalent to an ndarray
    call. So ``mt.block([[1, 2], [3, 4]])`` is equivalent to
    ``mt.array([[1, 2], [3, 4]])``.

    This function does not enforce that the blocks lie on a fixed grid.
    ``mt.block([[a, b], [c, d]])`` is not restricted to arrays of the form::

        AAAbb
        AAAbb
        cccDD

    But is also allowed to produce, for some ``a, b, c, d``::

        AAAbb
        AAAbb
        cDDDD

    Since concatenation happens along the last axis first, `block` is _not_
    capable of producing the following directly::

        AAAbb
        cccbb
        cccDD

    Matlab's "square bracket stacking", ``[A, B, ...; p, q, ...]``, is
    equivalent to ``mt.block([[A, B, ...], [p, q, ...]])``.

    Examples
    --------
    The most common use of this function is to build a block matrix

    >>> import mars.tensor as mt
    >>> A = mt.eye(2) * 2
    >>> B = mt.eye(3) * 3
    >>> mt.block([
    ...     [A,               mt.zeros((2, 3))],
    ...     [mt.ones((3, 2)), B               ]
    ... ]).execute()
    array([[2., 0., 0., 0., 0.],
           [0., 2., 0., 0., 0.],
           [1., 1., 3., 0., 0.],
           [1., 1., 0., 3., 0.],
           [1., 1., 0., 0., 3.]])

    With a list of depth 1, `block` can be used as `hstack`

    >>> mt.block([1, 2, 3]).execute()    # hstack([1, 2, 3])
    array([1, 2, 3])

    >>> a = mt.array([1, 2, 3])
    >>> b = mt.array([2, 3, 4])
    >>> mt.block([a, b, 10]).execute()   # hstack([a, b, 10])
    array([ 1,  2,  3,  2,  3,  4, 10])

    >>> A = mt.ones((2, 2), int)
    >>> B = 2 * A
    >>> mt.block([A, B]).execute()       # hstack([A, B])
    array([[1, 1, 2, 2],
           [1, 1, 2, 2]])

    With a list of depth 2, `block` can be used in place of `vstack`:

    >>> a = mt.array([1, 2, 3])
    >>> b = mt.array([2, 3, 4])
    >>> mt.block([[a], [b]]).execute()   # vstack([a, b])
    array([[1, 2, 3],
           [2, 3, 4]])

    >>> A = mt.ones((2, 2), int)
    >>> B = 2 * A
    >>> mt.block([[A], [B]]).execute()   # vstack([A, B])
    array([[1, 1],
           [1, 1],
           [2, 2],
           [2, 2]])

    It can also be used in places of `atleast_1d` and `atleast_2d`

    >>> a = mt.array(0)
    >>> b = mt.array([1])
    >>> mt.block([a]).execute()          # atleast_1d(a)
    array([0])
    >>> mt.block([b]).execute()          # atleast_1d(b)
    array([1])

    >>> mt.block([[a]]).execute()        # atleast_2d(a)
    array([[0]])
    >>> mt.block([[b]]).execute()        # atleast_2d(b)
    array([[1]])


    """
    arrays, list_ndim, result_ndim, final_size = _block_setup(arrays)

    # It was found through benchmarking that making an array of final size
    # around 256x256 was faster by straight concatenation on a
    # i7-7700HQ processor and dual channel ram 2400MHz.
    # It didn't seem to matter heavily on the dtype used.
    #
    # A 2D array using repeated concatenation requires 2 copies of the array.
    #
    # The fastest algorithm will depend on the ratio of CPU power to memory
    # speed.
    # One can monitor the results of the benchmark
    # https://pv.github.io/numpy-bench/#bench_shape_base.Block2D.time_block2d
    # to tune this parameter until a C version of the `_block_info_recursion`
    # algorithm is implemented which would likely be faster than the python
    # version.
    if list_ndim * final_size > (2 * 512 * 512):
        return _block_slicing(arrays, list_ndim, result_ndim)
    else:
        return _block_concatenate(arrays, list_ndim, result_ndim)


# These helper functions are mostly used for testing.
# They allow us to write tests that directly call `_block_slicing`
# or `_block_concatenate` without blocking large arrays to force the wisdom
# to trigger the desired path.
def _block_setup(arrays):
    """
    Returns
    (`arrays`, list_ndim, result_ndim, final_size)
    """
    bottom_index, arr_ndim, final_size = _block_check_depths_match(arrays)
    list_ndim = len(bottom_index)
    if bottom_index and bottom_index[-1] is None:
        raise ValueError(
            "List at {} cannot be empty".format(_block_format_index(bottom_index))
        )
    result_ndim = max(arr_ndim, list_ndim)
    return arrays, list_ndim, result_ndim, final_size


def _block_slicing(arrays, list_ndim, result_ndim):
    shape, slices, arrays = _block_info_recursion(arrays, list_ndim, result_ndim)
    dtype = np.result_type(*[arr.dtype for arr in arrays])

    # Test preferring F only in the case that all input arrays are F
    F_order = all(arr.flags["F_CONTIGUOUS"] for arr in arrays)
    C_order = all(arr.flags["C_CONTIGUOUS"] for arr in arrays)
    order = "F" if F_order and not C_order else "C"
    result = empty(shape=shape, dtype=dtype, order=order)
    # Note: In a c implementation, the function
    # PyArray_CreateMultiSortedStridePerm could be used for more advanced
    # guessing of the desired order.

    for the_slice, arr in zip(slices, arrays):
        result[(Ellipsis,) + the_slice] = arr
    return result


def _block_concatenate(arrays, list_ndim, result_ndim):
    result = _block(arrays, list_ndim, result_ndim)
    if list_ndim == 0:
        # Catch an edge case where _block returns a view because
        # `arrays` is a single mars array and not a list of mars arrays.
        # This might copy scalars or lists twice, but this isn't a likely
        # usecase for those interested in performance
        result = result.copy()
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import Int32Field
from ..array_utils import device, as_same_device
from ..datasource import tensor as astensor
from .core import TensorUnaryOp
from .utils import arithmetic_operand


@arithmetic_operand(init=False, sparse_mode="unary")
class TensorAround(TensorUnaryOp):
    _op_type_ = OperandDef.AROUND

    _decimals = Int32Field("decimals")
    _func_name = "around"

    @property
    def decimals(self):
        return self._decimals

    def __init__(
        self,
        decimals=None,
        casting="same_kind",
        err=None,
        dtype=None,
        sparse=False,
        **kw
    ):
        err = err if err is not None else np.geterr()
        super().__init__(
            _decimals=decimals,
            _casting=casting,
            _err=err,
            dtype=dtype,
            sparse=sparse,
            **kw
        )

    @property
    def ufunc_extra_params(self):
        return {"decimals": self._decimals}

    @classmethod
    def execute(cls, ctx, op):
        (a,), device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        with device(device_id):
            ctx[op.outputs[0].key] = xp.around(a, decimals=op.decimals)


def around(a, decimals=0, out=None):
    """
    Evenly round to the given number of decimals.

    Parameters
    ----------
    a : array_like
        Input data.
    decimals : int, optional
        Number of decimal places to round to (default: 0).  If
        decimals is negative, it specifies the number of positions to
        the left of the decimal point.
    out : Tensor, optional
        Alternative output tensor in which to place the result. It must have
        the same shape as the expected output, but the type of the output
        values will be cast if necessary.

    Returns
    -------
    rounded_array : Tensor
        An tensor of the same type as `a`, containing the rounded values.
        Unless `out` was specified, a new tensor is created.  A reference to
        the result is returned.

        The real and imaginary parts of complex numbers are rounded
        separately.  The result of rounding a float is a float.

    See Also
    --------
    Tensor.round : equivalent method

    ceil, fix, floor, rint, trunc


    Notes
    -----
    For values exactly halfway between rounded decimal values, NumPy
    rounds to the nearest even value. Thus 1.5 and 2.5 round to 2.0,
    -0.5 and 0.5 round to 0.0, etc. Results may also be surprising due
    to the inexact representation of decimal fractions in the IEEE
    floating point standard [1]_ and errors introduced when scaling
    by powers of ten.

    References
    ----------
    .. [1] "Lecture Notes on the Status of  IEEE 754", William Kahan,
           http://www.cs.berkeley.edu/~wkahan/ieee754status/IEEE754.PDF
    .. [2] "How Futile are Mindless Assessments of
           Roundoff in Floating-Point Computation?", William Kahan,
           http://www.cs.berkeley.edu/~wkahan/Mindless.pdf

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.around([0.37, 1.64]).execute()
    array([ 0.,  2.])
    >>> mt.around([0.37, 1.64], decimals=1).execute()
    array([ 0.4,  1.6])
    >>> mt.around([.5, 1.5, 2.5, 3.5, 4.5]).execute() # rounds to nearest even value
    array([ 0.,  2.,  2.,  4.,  4.])
    >>> mt.around([1,2,3,11], decimals=1).execute() # tensor of ints is returned
    array([ 1,  2,  3, 11])
    >>> mt.around([1,2,3,11], decimals=-1).execute()
    array([ 0,  0,  0, 10])

    """
    dtype = astensor(a).dtype
    op = TensorAround(decimals=decimals, dtype=dtype)
    return op(a, out=out)


round_ = around



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from ...backend import BaseActorBackend, register_backend
from ..context import MarsActorContext
from .driver import MarsActorDriver
from .pool import MainActorPool


__all__ = ["MarsActorBackend"]


def build_pool_kwargs(n_process: int, kwargs: Dict):
    n_io_process = kwargs.pop("n_io_process", 0)
    if n_io_process:
        n_process += n_io_process

        labels = kwargs["labels"]
        envs = kwargs["envs"]
        # sub-pools for IO(transfer and spill)
        for _ in range(n_io_process):
            if envs:  # pragma: no cover
                envs.append(dict())
            labels.append("io")
    return n_process, kwargs


@register_backend
class MarsActorBackend(BaseActorBackend):
    @staticmethod
    def name():
        # return None because Mars is default scheme
        return

    @staticmethod
    def get_context_cls():
        return MarsActorContext

    @staticmethod
    def get_driver_cls():
        return MarsActorDriver

    @classmethod
    async def create_actor_pool(cls, address: str, n_process: int = None, **kwargs):
        from ..pool import create_actor_pool

        n_process, kwargs = build_pool_kwargs(n_process, kwargs)
        return await create_actor_pool(
            address, pool_cls=MainActorPool, n_process=n_process, **kwargs
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...core.context import get_context
from ...core.custom_log import redirect_custom_log
from ...serialization.serializables import (
    BoolField,
    TupleField,
    DictField,
    FunctionField,
)
from ...core.operand import OperatorLogicKeyGeneratorMixin
from ...utils import enter_current_session, quiet_stdio, get_func_token_values
from ..operands import DataFrameOperandMixin, DataFrameOperand
from ..utils import (
    auto_merge_chunks,
    build_empty_df,
    build_empty_series,
    parse_index,
    validate_output_types,
    make_dtypes,
    make_dtype,
)


class GroupByApplyLogicKeyGeneratorMixin(OperatorLogicKeyGeneratorMixin):
    def _get_logic_key_token_values(self):
        token_values = super()._get_logic_key_token_values()
        if self.func:
            return token_values + get_func_token_values(self.func)
        else:  # pragma: no cover
            return token_values


class GroupByApply(
    DataFrameOperand, DataFrameOperandMixin, GroupByApplyLogicKeyGeneratorMixin
):
    _op_type_ = opcodes.APPLY
    _op_module_ = "dataframe.groupby"

    func = FunctionField("func")
    args = TupleField("args", default_factory=tuple)
    kwds = DictField("kwds", default_factory=dict)
    maybe_agg = BoolField("maybe_agg", default=None)

    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op):
        in_data = ctx[op.inputs[0].key]
        out = op.outputs[0]
        if not in_data:
            if op.output_types[0] == OutputType.dataframe:
                ctx[op.outputs[0].key] = build_empty_df(op.outputs[0].dtypes)
            else:
                ctx[op.outputs[0].key] = build_empty_series(
                    op.outputs[0].dtype, name=out.name
                )
            return

        applied = in_data.apply(op.func, *op.args, **op.kwds)

        if isinstance(applied, pd.DataFrame):
            # when there is only one group, pandas tend to return a DataFrame, while
            # we need to convert it into a compatible series
            if op.output_types[0] == OutputType.series:
                assert len(applied.index) == 1
                applied_idx = pd.MultiIndex.from_arrays(
                    [
                        [applied.index[0]] * len(applied.columns),
                        applied.columns.tolist(),
                    ]
                )
                applied_idx.names = [applied.index.name, None]
                applied = pd.Series(
                    np.array(applied.iloc[0]), applied_idx, name=applied.columns.name
                )
            else:
                applied.columns.name = None
        else:
            applied.name = out.name
        ctx[out.key] = applied

    @classmethod
    def tile(cls, op):
        in_groupby = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_groupby.chunks:
            inp_chunks = [c]

            new_op = op.copy().reset_key()
            new_op.tileable_op_key = op.key
            if op.output_types[0] == OutputType.dataframe:
                chunks.append(
                    new_op.new_chunk(
                        inp_chunks,
                        index=c.index,
                        shape=(np.nan, len(out_df.dtypes)),
                        dtypes=out_df.dtypes,
                        columns_value=out_df.columns_value,
                        index_value=out_df.index_value,
                    )
                )
            else:
                chunks.append(
                    new_op.new_chunk(
                        inp_chunks,
                        name=out_df.name,
                        index=(c.index[0],),
                        shape=(np.nan,),
                        dtype=out_df.dtype,
                        index_value=out_df.index_value,
                    )
                )

        new_op = op.copy()
        kw = out_df.params.copy()
        kw["chunks"] = chunks
        if op.output_types[0] == OutputType.dataframe:
            kw["nsplits"] = ((np.nan,) * len(chunks), (out_df.shape[1],))
        else:
            kw["nsplits"] = ((np.nan,) * len(chunks),)
        ret = new_op.new_tileable([in_groupby], **kw)
        if not op.maybe_agg:
            return [ret]
        else:
            # auto merge small chunks if df.groupby().apply(func)
            # may be an aggregation operation
            yield ret.chunks  # trigger execution for chunks
            return [auto_merge_chunks(get_context(), ret)]

    def _infer_df_func_returns(
        self, in_groupby, in_df, dtypes=None, dtype=None, name=None, index=None
    ):
        index_value, output_type, new_dtypes = None, None, None

        if self.output_types is not None and (dtypes is not None or dtype is not None):
            ret_dtypes = dtypes if dtypes is not None else (dtype, name)
            ret_index_value = parse_index(index) if index is not None else None
            return ret_dtypes, ret_index_value

        try:
            infer_df = in_groupby.op.build_mock_groupby().apply(
                self.func, *self.args, **self.kwds
            )

            if len(infer_df) <= 2:
                # we create mock df with 4 rows, 2 groups
                # if return df has 2 rows, we assume that
                # it's an aggregation operation
                self.maybe_agg = True

            # todo return proper index when sort=True is implemented
            index_value = parse_index(infer_df.index[:0], in_df.key, self.func)

            # for backward compatibility
            dtype = dtype if dtype is not None else dtypes
            if isinstance(infer_df, pd.DataFrame):
                output_type = output_type or OutputType.dataframe
                new_dtypes = new_dtypes or infer_df.dtypes
            elif isinstance(infer_df, pd.Series):
                output_type = output_type or OutputType.series
                new_dtypes = new_dtypes or (
                    name or infer_df.name,
                    dtype or infer_df.dtype,
                )
            else:
                output_type = OutputType.series
                new_dtypes = (name, dtype or pd.Series(infer_df).dtype)
        except:  # noqa: E722  # nosec
            pass

        self.output_types = (
            [output_type] if not self.output_types else self.output_types
        )
        dtypes = new_dtypes if dtypes is None else dtypes
        index_value = index_value if index is None else parse_index(index)
        return dtypes, index_value

    def __call__(self, groupby, dtypes=None, dtype=None, name=None, index=None):
        in_df = groupby
        while in_df.op.output_types[0] not in (OutputType.dataframe, OutputType.series):
            in_df = in_df.inputs[0]

        with quiet_stdio():
            dtypes, index_value = self._infer_df_func_returns(
                groupby, in_df, dtypes, dtype=dtype, name=name, index=index
            )
        if index_value is None:
            index_value = parse_index(None, (in_df.key, in_df.index_value.key))
        for arg, desc in zip((self.output_types, dtypes), ("output_types", "dtypes")):
            if arg is None:
                raise TypeError(
                    f"Cannot determine {desc} by calculating with enumerate data, "
                    "please specify it as arguments"
                )

        if self.output_types[0] == OutputType.dataframe:
            new_shape = (np.nan, len(dtypes))
            return self.new_dataframe(
                [groupby],
                shape=new_shape,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=parse_index(dtypes.index, store_data=True),
            )
        else:
            name = name or dtypes[0]
            dtype = dtype or dtypes[1]
            new_shape = (np.nan,)
            return self.new_series(
                [groupby],
                name=name,
                shape=new_shape,
                dtype=dtype,
                index_value=index_value,
            )


def groupby_apply(
    groupby,
    func,
    *args,
    output_type=None,
    dtypes=None,
    dtype=None,
    name=None,
    index=None,
    **kwargs,
):
    """
    Apply function `func` group-wise and combine the results together.

    The function passed to `apply` must take a dataframe as its first
    argument and return a DataFrame, Series or scalar. `apply` will
    then take care of combining the results back together into a single
    dataframe or series. `apply` is therefore a highly flexible
    grouping method.

    While `apply` is a very flexible method, its downside is that
    using it can be quite a bit slower than using more specific methods
    like `agg` or `transform`. Pandas offers a wide range of method that will
    be much faster than using `apply` for their specific purposes, so try to
    use them before reaching for `apply`.

    Parameters
    ----------
    func : callable
        A callable that takes a dataframe as its first argument, and
        returns a dataframe, a series or a scalar. In addition the
        callable may take positional and keyword arguments.

    output_type : {'dataframe', 'series'}, default None
        Specify type of returned object. See `Notes` for more details.

    dtypes : Series, default None
        Specify dtypes of returned DataFrames. See `Notes` for more details.

    dtype : numpy.dtype, default None
        Specify dtype of returned Series. See `Notes` for more details.

    name : str, default None
        Specify name of returned Series. See `Notes` for more details.

    index : Index, default None
        Specify index of returned object. See `Notes` for more details.

    args, kwargs : tuple and dict
        Optional positional and keyword arguments to pass to `func`.

    Returns
    -------
    applied : Series or DataFrame

    See Also
    --------
    pipe : Apply function to the full GroupBy object instead of to each
        group.
    aggregate : Apply aggregate function to the GroupBy object.
    transform : Apply function column-by-column to the GroupBy object.
    Series.apply : Apply a function to a Series.
    DataFrame.apply : Apply a function to each row or column of a DataFrame.

    Notes
    -----
    When deciding output dtypes and shape of the return value, Mars will
    try applying ``func`` onto a mock grouped object, and the apply call
    may fail. When this happens, you need to specify the type of apply
    call (DataFrame or Series) in output_type.

    * For DataFrame output, you need to specify a list or a pandas Series
      as ``dtypes`` of output DataFrame. ``index`` of output can also be
      specified.
    * For Series output, you need to specify ``dtype`` and ``name`` of
      output Series.
    """
    output_types = kwargs.pop("output_types", None)
    object_type = kwargs.pop("object_type", None)
    output_types = validate_output_types(
        output_types=output_types, output_type=output_type, object_type=object_type
    )

    dtypes = make_dtypes(dtypes)
    dtype = make_dtype(dtype)
    op = GroupByApply(func=func, args=args, kwds=kwargs, output_types=output_types)
    return op(groupby, dtypes=dtypes, dtype=dtype, name=name, index=index)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .power_divergence import power_divergence


def chisquare(f_obs, f_exp=None, ddof=0, axis=0):
    """
    Calculate a one-way chi-square test.

    The chi-square test tests the null hypothesis that the categorical data
    has the given frequencies.

    Parameters
    ----------
    f_obs : array_like
        Observed frequencies in each category.
    f_exp : array_like, optional
        Expected frequencies in each category.  By default the categories are
        assumed to be equally likely.
    ddof : int, optional
        "Delta degrees of freedom": adjustment to the degrees of freedom
        for the p-value.  The p-value is computed using a chi-squared
        distribution with ``k - 1 - ddof`` degrees of freedom, where `k`
        is the number of observed frequencies.  The default value of `ddof`
        is 0.
    axis : int or None, optional
        The axis of the broadcast result of `f_obs` and `f_exp` along which to
        apply the test.  If axis is None, all values in `f_obs` are treated
        as a single data set.  Default is 0.

    Returns
    -------
    chisq : float or ndarray
        The chi-squared test statistic.  The value is a float if `axis` is
        None or `f_obs` and `f_exp` are 1-D.
    p : float or ndarray
        The p-value of the test.  The value is a float if `ddof` and the
        return value `chisq` are scalars.

    See Also
    --------
    scipy.stats.power_divergence

    Notes
    -----
    This test is invalid when the observed or expected frequencies in each
    category are too small.  A typical rule is that all of the observed
    and expected frequencies should be at least 5.

    The default degrees of freedom, k-1, are for the case when no parameters
    of the distribution are estimated. If p parameters are estimated by
    efficient maximum likelihood then the correct degrees of freedom are
    k-1-p. If the parameters are estimated in a different way, then the
    dof can be between k-1-p and k-1. However, it is also possible that
    the asymptotic distribution is not chi-square, in which case this test
    is not appropriate.

    References
    ----------
    .. [1] Lowry, Richard.  "Concepts and Applications of Inferential
           Statistics". Chapter 8.
           https://web.archive.org/web/20171022032306/http://vassarstats.net:80/textbook/ch8pt1.html
    .. [2] "Chi-squared test", https://en.wikipedia.org/wiki/Chi-squared_test

    Examples
    --------
    When just `f_obs` is given, it is assumed that the expected frequencies
    are uniform and given by the mean of the observed frequencies.

    >>> import mars.tensor as mt
    >>> from mars.tensor.stats import chisquare
    >>> chisquare([16, 18, 16, 14, 12, 12])
    (2.0, 0.84914503608460956)

    With `f_exp` the expected frequencies can be given.

    >>> chisquare([16, 18, 16, 14, 12, 12], f_exp=[16, 16, 16, 16, 16, 8]).execute()
    (3.5, 0.62338762774958223)

    When `f_obs` is 2-D, by default the test is applied to each column.

    >>> obs = mt.array([[16, 18, 16, 14, 12, 12], [32, 24, 16, 28, 20, 24]]).T
    >>> obs.shape
    (6, 2)
    >>> chisquare(obs).execute()
    (array([ 2.        ,  6.66666667]), array([ 0.84914504,  0.24663415]))

    By setting ``axis=None``, the test is applied to all data in the array,
    which is equivalent to applying the test to the flattened array.

    >>> chisquare(obs, axis=None).execute()
    (23.31034482758621, 0.015975692534127565)
    >>> chisquare(obs.ravel()).execute()
    (23.31034482758621, 0.015975692534127565)

    `ddof` is the change to make to the default degrees of freedom.

    >>> chisquare([16, 18, 16, 14, 12, 12], ddof=1).execute()
    (2.0, 0.73575888234288467)

    `f_obs` and `f_exp` are also broadcast.  In the following, `f_obs` has
    shape (6,) and `f_exp` has shape (2, 6), so the result of broadcasting
    `f_obs` and `f_exp` has shape (2, 6).  To compute the desired chi-squared
    statistics, we use ``axis=1``:

    >>> chisquare([16, 18, 16, 14, 12, 12],
    ...           f_exp=[[16, 16, 16, 16, 16, 8], [8, 20, 20, 16, 12, 12]],
    ...           axis=1).execute()
    (array([ 3.5 ,  9.25]), array([ 0.62338763,  0.09949846]))

    """
    return power_divergence(f_obs, f_exp=f_exp, ddof=ddof, axis=axis, lambda_="pearson")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorFFTShiftMixin, TensorFFTShiftBase


class TensorIFFTShift(TensorFFTShiftBase, TensorFFTShiftMixin):
    _op_type_ = OperandDef.IFFTSHIFT

    def __init__(self, axes=None, **kw):
        super().__init__(_axes=axes, **kw)

    @classmethod
    def _is_inverse(cls):
        return True

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, x):
        return self.new_tensor([x], x.shape)


def ifftshift(x, axes=None):
    """
    The inverse of `fftshift`. Although identical for even-length `x`, the
    functions differ by one sample for odd-length `x`.

    Parameters
    ----------
    x : array_like
        Input tensor.
    axes : int or shape tuple, optional
        Axes over which to calculate.  Defaults to None, which shifts all axes.

    Returns
    -------
    y : Tensor
        The shifted tensor.

    See Also
    --------
    fftshift : Shift zero-frequency component to the center of the spectrum.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> freqs = mt.fft.fftfreq(9, d=1./9).reshape(3, 3)
    >>> freqs.execute()
    array([[ 0.,  1.,  2.],
           [ 3.,  4., -4.],
           [-3., -2., -1.]])
    >>> mt.fft.ifftshift(mt.fft.fftshift(freqs)).execute()
    array([[ 0.,  1.,  2.],
           [ 3.,  4., -4.],
           [-3., -2., -1.]])

    """
    x = astensor(x)
    dtype = np.fft.ifftshift(np.empty((1,) * max(1, x.ndim), dtype=x.dtype)).dtype
    axes = TensorIFFTShift._process_axes(x, axes)
    op = TensorIFFTShift(axes=axes, dtype=dtype)
    return op(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, OutputType, recursive_tile
from ...core.context import Context
from ...serialization.serializables import KeyField, AnyField
from ...tensor.core import Tensor
from ...tensor.datasource import tensor as astensor
from ...tensor.utils import unify_chunks
from ...typing import EntityType, TileableType
from ...utils import has_unknown_shape
from ..core import INDEX_TYPE, SERIES_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index


class DataFrameFromTensor(DataFrameOperand, DataFrameOperandMixin):
    """
    Represents data from mars tensor
    """

    _op_type_ = OperandDef.DATAFRAME_FROM_TENSOR

    input = AnyField("input")
    index = AnyField("index")
    columns = AnyField("columns")

    def __init__(self, *args, **kwargs):
        kwargs["_output_types"] = [OutputType.dataframe]
        super().__init__(*args, **kwargs)

    def _set_inputs(self, inputs: List[EntityType]):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        if self.input is not None:
            if not isinstance(self.input, dict):
                self.input = next(inputs_iter)
            else:
                # check each value for input
                new_input = OrderedDict()
                for k, v in self.input.items():
                    if isinstance(v, ENTITY_TYPE):
                        new_input[k] = next(inputs_iter)
                    else:
                        new_input[k] = v
                self.input = new_input

        if isinstance(self.index, ENTITY_TYPE):
            self.index = next(inputs_iter)

    def __call__(
        self,
        input_tensor: Tensor,
        index: Union[TileableType, pd.Index],
        columns: pd.Index,
        dtypes: pd.Series,
    ):
        if isinstance(input_tensor, dict):
            return self._call_input_1d_tileables(input_tensor, index, columns, dtypes)
        elif input_tensor is not None:
            return self._call_input_tensor(input_tensor, index, columns, dtypes)
        else:
            return self._call_tensor_none(index, columns, dtypes)

    def _process_index(
        self, index: Union[TileableType, pd.Index], inputs: List[EntityType]
    ):
        if not isinstance(index, pd.Index):
            if isinstance(index, INDEX_TYPE):
                index_value = index.index_value
                inputs.append(index)
            elif isinstance(index, ENTITY_TYPE):
                index = astensor(index)
                if index.ndim != 1:
                    raise ValueError(f"index should be 1-d, got {index.ndim}-d")
                index_value = parse_index(
                    pd.Index([], dtype=index.dtype), index, type(self).__name__
                )
                inputs.append(index)
            else:
                index = pd.Index(index)
                index_value = parse_index(index)
        else:
            index_value = parse_index(index)
        return index_value

    def _call_input_1d_tileables(
        self,
        input_1d_tileables: Dict[Any, TileableType],
        index: Union[TileableType, pd.Index],
        columns: pd.Index,
        dtypes: pd.Series,
    ):
        tileables = []
        shape = None
        for tileable in input_1d_tileables.values():
            tileable_shape = astensor(tileable).shape
            if len(tileable_shape) > 0:
                if shape is None:
                    shape = tileable_shape
                elif shape != tileable_shape:
                    raise ValueError("input 1-d tensors should have same shape")

            if isinstance(tileable, ENTITY_TYPE):
                tileables.append(tileable)

        if index is not None:
            tileable_size = tileables[0].shape[0]
            if hasattr(index, "shape"):
                index_size = index.shape[0]
            else:
                index_size = len(index)
            if (
                not pd.isna(tileable_size)
                and not pd.isna(index_size)
                and tileable_size != index_size
            ):
                raise ValueError(
                    f"index {index} should have the same shape "
                    f"with tensor: {tileable_size}"
                )
            index_value = self._process_index(index, tileables)
        else:
            self.index = index = pd.RangeIndex(0, tileables[0].shape[0])
            index_value = parse_index(index)

        if columns is not None:
            if len(input_1d_tileables) != len(columns):
                raise ValueError(
                    f"columns {columns} should have size {len(input_1d_tileables)}"
                )
            if not isinstance(columns, pd.Index):
                if isinstance(columns, ENTITY_TYPE):
                    raise NotImplementedError("The columns value cannot be a tileable")
                columns = pd.Index(columns)
            columns_value = parse_index(columns, store_data=True)
        else:
            columns_value = parse_index(
                pd.RangeIndex(0, len(input_1d_tileables)), store_data=True
            )

        shape = (shape[0], len(input_1d_tileables))
        return self.new_dataframe(
            tileables,
            shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    def _call_input_tensor(
        self,
        input_tensor: Tensor,
        index: Union[TileableType, pd.Index],
        columns: pd.Index,
        dtypes: pd.Series,
    ):
        if input_tensor.ndim not in {1, 2}:
            raise ValueError("Must pass 1-d or 2-d input")
        inputs = [input_tensor]

        if index is not None:
            if input_tensor.shape[0] != len(index):
                raise ValueError(
                    f"index {index} should have the same shape with tensor: {input_tensor.shape[0]}"
                )
            index_value = self._process_index(index, inputs)
        elif isinstance(input_tensor, SERIES_TYPE):
            index_value = input_tensor.index_value
        else:
            stop = input_tensor.shape[0]
            stop = -1 if np.isnan(stop) else stop
            index = self.index = pd.RangeIndex(start=0, stop=stop)
            index_value = parse_index(index)

        if columns is not None:
            if not (
                input_tensor.ndim == 1
                and len(columns) == 1
                or input_tensor.shape[1] == len(columns)
            ):
                raise ValueError(
                    f"columns {columns} should have the same shape with tensor: {input_tensor.shape[1]}"
                )
            if not isinstance(columns, pd.Index):
                if isinstance(columns, ENTITY_TYPE):
                    raise NotImplementedError("The columns value cannot be a tileable")
                columns = pd.Index(columns)
            columns_value = parse_index(columns, store_data=True)
        else:
            if input_tensor.ndim == 1:
                # convert to 1-d DataFrame
                columns_value = parse_index(
                    pd.RangeIndex(start=0, stop=1), store_data=True
                )
            else:
                columns_value = parse_index(
                    pd.RangeIndex(start=0, stop=input_tensor.shape[1]), store_data=True
                )

        if input_tensor.ndim == 1:
            shape = (input_tensor.shape[0], 1)
        else:
            shape = input_tensor.shape

        return self.new_dataframe(
            inputs,
            shape,
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    def _call_tensor_none(
        self, index: Union[TileableType, pd.Index], columns: pd.Index, dtypes: pd.Series
    ):
        inputs = []
        shape = []
        if index is not None:
            index_value = self._process_index(index, inputs)
            shape.append(index.shape[0])
        else:
            index = self.index = pd.Index([], dtype=object)
            index_value = parse_index(index)
            shape.append(0)

        if columns is not None:
            if not isinstance(columns, pd.Index):
                if isinstance(columns, ENTITY_TYPE):
                    raise NotImplementedError("The columns value cannot be a tileable")
                columns = pd.Index(columns)
            columns_value = parse_index(columns, store_data=True)
            shape.append(columns.shape[0])
        else:
            columns_value = parse_index(pd.Index([], dtype=object), store_data=True)
            shape.append(0)

        return self.new_dataframe(
            inputs,
            shape=tuple(shape),
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )

    @classmethod
    def tile(cls, op: "DataFrameFromTensor"):
        if isinstance(op.input, dict):
            return (yield from cls._tile_input_1d_tileables(op))
        elif op.input is not None:
            return (yield from cls._tile_input_tensor(op))
        else:
            return cls._tile_tensor_none(op)

    @classmethod
    def _tile_input_1d_tileables(cls, op: "DataFrameFromTensor"):
        # make sure all tensor have known chunk shapes
        if has_unknown_shape(*op.inputs):
            yield

        out_df = op.outputs[0]
        in_tensors = op.inputs
        in_tensors = yield from unify_chunks(*in_tensors)
        nsplit = in_tensors[0].nsplits[0]

        cum_sizes = [0] + np.cumsum(nsplit).tolist()
        out_chunks = []
        for i in range(in_tensors[0].chunk_shape[0]):
            chunk_op = op.copy().reset_key()
            new_input = OrderedDict()
            for k, v in op.input.items():
                if not isinstance(v, ENTITY_TYPE):
                    try:
                        new_input[k] = v[cum_sizes[i] : cum_sizes[i + 1]]
                    except TypeError:
                        # scalar
                        new_input[k] = v
                else:
                    # do not need to do slice,
                    # will be done in set_inputs
                    new_input[k] = v
            chunk_op.input = new_input
            columns_value = out_df.columns_value
            dtypes = out_df.dtypes
            chunk_index = (i, 0)
            if isinstance(op.index, INDEX_TYPE):
                index_value = in_tensors[-1].chunks[i].index_value
            elif isinstance(op.index, pd.Index):
                chunk_op.index = pd_index = op.index[cum_sizes[i] : cum_sizes[i + 1]]
                index_value = parse_index(pd_index, store_data=True)
            else:
                assert op.index is not None
                index_chunk = in_tensors[-1].cix[
                    i,
                ]
                index_value = parse_index(
                    pd.Index([], dtype=index_chunk.dtype),
                    index_chunk,
                    type(chunk_op).__name__,
                )
            shape = (nsplit[i], len(out_df.dtypes))
            out_chunk = chunk_op.new_chunk(
                [t.cix[(i,)] for t in in_tensors],
                shape=shape,
                index=chunk_index,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )
            out_chunks.append(out_chunk)

        nsplits = (nsplit, (len(out_df.dtypes),))
        new_op = op.copy()
        return new_op.new_dataframes(
            out_df.inputs,
            out_df.shape,
            dtypes=out_df.dtypes,
            index_value=out_df.index_value,
            columns_value=out_df.columns_value,
            chunks=out_chunks,
            nsplits=nsplits,
        )

    @classmethod
    def _tile_input_tensor(cls, op: "DataFrameFromTensor"):
        out_df = op.outputs[0]
        in_tensor = op.input
        out_chunks = []
        if out_df.index_value.has_value() and has_unknown_shape(in_tensor):
            yield

        nsplits = in_tensor.nsplits

        if op.index is not None and hasattr(op.index, "key"):
            # rechunk index if it's a tensor
            if has_unknown_shape(*op.inputs):
                yield
            index_tensor = yield from recursive_tile(op.index.rechunk([nsplits[0]]))
        else:
            index_tensor = None

        # nsplits
        if in_tensor.ndim == 1:
            out_nsplits = in_tensor.nsplits + ((1,),)
        else:
            out_nsplits = in_tensor.nsplits

        cum_nsplits = [[0] + np.cumsum(ns).tolist() for ns in out_nsplits]
        for in_chunk in in_tensor.chunks:
            out_op = op.copy().reset_key()
            chunk_inputs = [in_chunk]
            if in_chunk.ndim == 1:
                i = in_chunk.index[0]
                chunk_index = (i, 0)
                chunk_shape = (in_chunk.shape[0], 1)
            else:
                i, j = in_chunk.index
                chunk_index = in_chunk.index
                chunk_shape = in_chunk.shape

            if op.columns is not None:
                column_nsplit = cum_nsplits[1]
                j = chunk_index[1]
                out_op.columns = op.columns[column_nsplit[j] : column_nsplit[j + 1]]

            if isinstance(op.index, INDEX_TYPE):
                index_chunk = index_tensor.chunks[i]
                chunk_inputs.append(index_chunk)
            elif isinstance(op.index, pd.Index):
                index_nsplit = cum_nsplits[0]
                if op.index.size > 0:
                    out_op.index = op.index[index_nsplit[i] : index_nsplit[i + 1]]
            elif index_tensor is not None:
                index_chunk = index_tensor.cix[i]
                chunk_inputs.append(index_chunk)

            out_chunk = out_op.new_chunk(
                chunk_inputs, shape=chunk_shape, index=chunk_index
            )
            out_chunk._set_tileable_meta(
                tileable_key=out_df.key,
                nsplits=out_nsplits,
                index_value=out_df.index_value,
                columns_value=out_df.columns_value,
                dtypes=out_df.dtypes,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_df.params.copy()
        params["chunks"] = out_chunks
        params["nsplits"] = out_nsplits
        return new_op.new_dataframes(out_df.inputs, kws=[params])

    @classmethod
    def _tile_tensor_none(cls, op: "DataFrameFromTensor"):
        out_df = op.outputs[0]

        out_chunks = []
        assert isinstance(op.index, INDEX_TYPE)
        # tile as index
        for index_chunk in op.index.chunks:
            index_value = index_chunk.index_value

            chunk_shape = (index_chunk.shape[0], out_df.shape[1])
            chunk_index = (index_chunk.index[0], 0)

            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk(
                [index_chunk],
                shape=chunk_shape,
                index=chunk_index,
                index_value=index_value,
                columns_value=out_df.columns_value,
                dtypes=out_df.dtypes,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_df.params.copy()
        params["nsplits"] = (op.index.nsplits[0], (out_df.shape[1],))
        params["chunks"] = out_chunks
        return new_op.new_dataframes(out_df.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "DataFrameFromTensor"):
        chunk = op.outputs[0]

        if isinstance(op.input, dict):
            d = OrderedDict()
            for k, v in op.input.items():
                if hasattr(v, "key"):
                    d[k] = ctx[v.key]
                else:
                    d[k] = v
            if op.index is not None and hasattr(op.index, "key"):
                index_data = ctx[op.index.key]
            else:
                index_data = op.index
            ctx[chunk.key] = pd.DataFrame(d, index=index_data, columns=op.columns)
        elif op.input is not None:
            tensor_data = ctx[op.inputs[0].key]
            if isinstance(tensor_data, pd.Series):
                ctx[chunk.key] = tensor_data.to_frame(name=chunk.dtypes.index[0])
            else:
                if op.index is not None and hasattr(op.index, "key"):
                    # index is a tensor
                    index_data = ctx[op.inputs[1].key]
                else:
                    index_data = op.index
                    if isinstance(index_data, pd.RangeIndex) and len(index_data) == 0:
                        index_data = None
                ctx[chunk.key] = pd.DataFrame(
                    tensor_data,
                    index=index_data,
                    columns=op.columns,
                )
        else:
            index_data = ctx[op.index.key]
            ctx[chunk.key] = pd.DataFrame(index=index_data, columns=op.columns)


def dataframe_from_tensor(
    tensor: Tensor,
    index: Union[TileableType, pd.Index] = None,
    columns: Union[pd.Index, list] = None,
    gpu: bool = None,
    sparse: bool = False,
):
    if tensor is not None:
        if tensor.ndim > 2 or tensor.ndim <= 0:
            raise TypeError(
                f"Not support create DataFrame from {tensor.ndim} dims tensor"
            )
        try:
            col_num = tensor.shape[1]
        except IndexError:
            col_num = 1
        gpu = tensor.op.gpu if gpu is None else gpu
        dtypes = pd.Series([tensor.dtype] * col_num, index=columns)
        if columns is None:
            columns = dtypes.index
    else:
        gpu = None
        if columns is not None:
            dtypes = pd.Series([], index=columns)
        else:
            dtypes = pd.Series([], index=pd.Index([], dtype=object))
    if index is not None and not isinstance(index, ENTITY_TYPE):
        index = pd.Index(index)
    op = DataFrameFromTensor(
        input=tensor, index=index, columns=columns, gpu=gpu, sparse=sparse
    )
    return op(tensor, index, columns, dtypes)


def dataframe_from_1d_tileables(
    d: Dict[Any, TileableType],
    index: Union[TileableType, pd.Index, list] = None,
    columns: Union[pd.Index, list] = None,
    gpu: bool = None,
    sparse: bool = False,
):
    data = dict()
    for k, v in d.items():
        if isinstance(v, (list, tuple)) and any(
            isinstance(sv, ENTITY_TYPE) for sv in v
        ):
            data[k] = astensor(v)
        else:
            data[k] = v
    d = data
    if columns is not None:
        tileables = [d.get(c) for c in columns]
    else:
        columns = list(d.keys())
        tileables = list(d.values())

    gpu = (
        next((t.op.gpu for t in tileables if hasattr(t, "op")), False)
        if gpu is None
        else gpu
    )
    dtypes = pd.Series(
        [t.dtype if hasattr(t, "dtype") else pd.Series(t).dtype for t in tileables],
        index=columns,
    )
    if index is not None and not isinstance(index, ENTITY_TYPE):
        index = pd.Index(index)
    op = DataFrameFromTensor(
        input=d, index=index, columns=columns, gpu=gpu, sparse=sparse
    )
    return op(d, index, columns, dtypes)


class SeriesFromTensor(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.SERIES_FROM_TENSOR

    input = KeyField("input")
    index = AnyField("index")

    def _set_inputs(self, inputs: List[EntityType]):
        super()._set_inputs(inputs)
        if self.input is not None:
            self.input = self.inputs[0]
        if self.index is not None and hasattr(self.index, "key"):
            self.index = self.inputs[-1]

    @classmethod
    def tile(cls, op: "SeriesFromTensor"):
        if op.index is None or not hasattr(op.index, "key"):
            # check all inputs to make sure no unknown chunk shape
            if has_unknown_shape(*op.inputs):
                yield

        if op.input is None:
            return cls._tile_tensor_none(op)

        out_series = op.outputs[0]
        in_tensor = op.inputs[0]
        nsplits = in_tensor.nsplits

        index_tensor = series_index = None
        if op.index is not None:
            if hasattr(op.index, "key"):
                index_tensor = yield from recursive_tile(op.index.rechunk([nsplits[0]]))
            else:
                series_index = op.index

        index_start = 0
        out_chunks = []
        for in_chunk in in_tensor.chunks:
            new_op = op.copy().reset_key()
            new_op.extra_params["index_start"] = index_start
            chunk_inputs = [in_chunk]
            if index_tensor is not None:
                index_chunk = index_tensor.cix[in_chunk.index]
                chunk_inputs.append(index_chunk)
                if isinstance(op.index, INDEX_TYPE):
                    index_value = index_chunk.index_value
                else:
                    index_value = parse_index(
                        pd.Index([], dtype=in_chunk.dtype),
                        index_chunk,
                        type(new_op).__name__,
                    )
            else:
                chunk_pd_index = series_index[
                    index_start : index_start + in_chunk.shape[0]
                ]
                index_value = parse_index(chunk_pd_index)
                new_op.index = chunk_pd_index
            index_start += in_chunk.shape[0]
            out_chunk = new_op.new_chunk(
                chunk_inputs,
                shape=in_chunk.shape,
                index=in_chunk.index,
                index_value=index_value,
                name=out_series.name,
                dtype=out_series.dtype,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs,
            shape=out_series.shape,
            dtype=out_series.dtype,
            index_value=out_series.index_value,
            name=out_series.name,
            chunks=out_chunks,
            nsplits=in_tensor.nsplits,
        )

    @classmethod
    def _tile_tensor_none(cls, op: "SeriesFromTensor"):
        out_series = op.outputs[0]

        out_chunks = []
        assert isinstance(op.index, INDEX_TYPE)
        # tile as index
        for index_chunk in op.index.chunks:
            index_value = index_chunk.index_value

            chunk_shape = (index_chunk.shape[0],)
            chunk_index = (index_chunk.index[0],)

            chunk_op = op.copy().reset_key()
            out_chunk = chunk_op.new_chunk(
                [index_chunk],
                shape=chunk_shape,
                index=chunk_index,
                index_value=index_value,
                dtype=out_series.dtype,
                name=out_series.name,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_series.params.copy()
        params["nsplits"] = (op.index.nsplits[0],)
        params["chunks"] = out_chunks
        return new_op.new_tileables(out_series.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "SeriesFromTensor"):
        chunk = op.outputs[0]
        if op.input is not None:
            tensor_data = ctx[op.input.key]
        else:
            tensor_data = None

        if op.index is not None and hasattr(op.index, "key"):
            index_data = ctx[op.index.key]
        else:
            index_data = op.index
            if (
                tensor_data is not None
                and isinstance(index_data, pd.RangeIndex)
                and len(index_data) == 0
            ):
                # index not specified
                index_data = None

        ctx[chunk.key] = pd.Series(
            tensor_data, index=index_data, name=chunk.name, dtype=chunk.dtype
        )

    def __call__(
        self,
        input_tensor: Tensor,
        index: Union[TileableType, pd.Index],
        dtype: np.dtype,
        name: Any,
    ):
        inputs = [input_tensor] if input_tensor is not None else []
        if index is not None:
            if not isinstance(index, pd.Index):
                if isinstance(index, INDEX_TYPE):
                    self.index = index
                    index_value = index.index_value
                    inputs.append(index)
                elif isinstance(index, ENTITY_TYPE):
                    self.index = index
                    index = astensor(index)
                    if index.ndim != 1:
                        raise ValueError(f"index should be 1-d, got {index.ndim}-d")
                    index_value = parse_index(
                        pd.Index([], dtype=index.dtype), index, type(self).__name__
                    )
                    inputs.append(index)
                else:
                    self.index = index = pd.Index(index)
                    index_value = parse_index(index)
            else:
                self.index = index
                index_value = parse_index(index)
        elif input_tensor is not None:
            if pd.isna(input_tensor.shape[0]):
                pd_index = pd.RangeIndex(-1)
            else:
                pd_index = pd.RangeIndex(start=0, stop=input_tensor.shape[0])
            index_value = parse_index(pd_index)
            self.index = pd_index
        else:
            self.index = index = pd.Index([], dtype=object)
            index_value = parse_index(index)

        if input_tensor is not None:
            shape = input_tensor.shape
        elif index is not None:
            shape = index.shape
        else:
            shape = (0,)

        return self.new_series(
            inputs, shape=shape, dtype=dtype, index_value=index_value, name=name
        )


def series_from_tensor(
    tensor: Tensor,
    index: Union[TileableType, pd.Index, list] = None,
    name: Any = None,
    dtype: np.dtype = None,
    gpu: bool = None,
    sparse: bool = False,
):
    if tensor is not None:
        if tensor.ndim > 1 or tensor.ndim <= 0:
            raise TypeError(f"Not support create Series from {tensor.ndim} dims tensor")
        gpu = tensor.op.gpu if gpu is None else gpu
        dtype = dtype or tensor.dtype
    else:
        gpu = None
        dtype = dtype or np.dtype(float)
    op = SeriesFromTensor(input=tensor, gpu=gpu, sparse=sparse)
    return op(tensor, index, dtype, name)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from dataclasses import dataclass
from typing import Tuple, Union, Type

from ...utils import to_binary, dataslots
from ..api import Actor
from ..core import ActorRef, create_local_actor_ref
from ..context import BaseActorContext
from ..debug import debug_async_timeout, detect_cycle_send
from ..errors import CannotCancelTask
from ..utils import create_actor_ref
from .allocate_strategy import AllocateStrategy, AddressSpecified
from .core import ActorCaller
from .message import (
    DEFAULT_PROTOCOL,
    new_message_id,
    _MessageBase,
    ResultMessage,
    ErrorMessage,
    CreateActorMessage,
    HasActorMessage,
    DestroyActorMessage,
    ActorRefMessage,
    SendMessage,
    CancelMessage,
    ControlMessage,
    ControlMessageType,
)
from .router import Router


@dataslots
@dataclass
class ProfilingContext:
    task_id: str


class MarsActorContext(BaseActorContext):
    __slots__ = ("_caller",)

    support_allocate_strategy = True

    def __init__(self, address: str = None):
        BaseActorContext.__init__(self, address)
        self._caller = ActorCaller()

    def __del__(self):
        self._caller.cancel_tasks()

    async def _call(
        self, address: str, message: _MessageBase, wait: bool = True
    ) -> Union[ResultMessage, ErrorMessage, asyncio.Future]:
        return await self._caller.call(
            Router.get_instance_or_empty(), address, message, wait=wait
        )

    @staticmethod
    def _process_result_message(message: Union[ResultMessage, ErrorMessage]):
        if isinstance(message, ResultMessage):
            return message.result
        else:
            raise message.as_instanceof_cause()

    async def _wait(self, future: asyncio.Future, address: str, message: _MessageBase):
        try:
            await asyncio.shield(future)
        except asyncio.CancelledError:
            try:
                await self.cancel(address, message.message_id)
            except CannotCancelTask:
                # cancel failed, already finished
                raise asyncio.CancelledError
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except
            pass
        return await future

    async def create_actor(
        self, actor_cls: Type[Actor], *args, uid=None, address: str = None, **kwargs
    ) -> ActorRef:
        router = Router.get_instance_or_empty()
        address = address or self._address or router.external_address
        allocate_strategy = kwargs.get("allocate_strategy", None)
        if isinstance(allocate_strategy, AllocateStrategy):
            allocate_strategy = kwargs.pop("allocate_strategy")
        else:
            allocate_strategy = AddressSpecified(address)
        create_actor_message = CreateActorMessage(
            new_message_id(),
            actor_cls,
            to_binary(uid),
            args,
            kwargs,
            allocate_strategy,
            protocol=DEFAULT_PROTOCOL,
        )
        future = await self._call(address, create_actor_message, wait=False)
        result = await self._wait(future, address, create_actor_message)
        return self._process_result_message(result)

    async def has_actor(self, actor_ref: ActorRef) -> bool:
        message = HasActorMessage(
            new_message_id(), actor_ref, protocol=DEFAULT_PROTOCOL
        )
        future = await self._call(actor_ref.address, message, wait=False)
        result = await self._wait(future, actor_ref.address, message)
        return self._process_result_message(result)

    async def destroy_actor(self, actor_ref: ActorRef):
        message = DestroyActorMessage(
            new_message_id(), actor_ref, protocol=DEFAULT_PROTOCOL
        )
        future = await self._call(actor_ref.address, message, wait=False)
        result = await self._wait(future, actor_ref.address, message)
        return self._process_result_message(result)

    async def kill_actor(self, actor_ref: ActorRef, force: bool = True):
        # get main_pool_address
        control_message = ControlMessage(
            new_message_id(),
            actor_ref.address,
            ControlMessageType.get_config,
            "main_pool_address",
            protocol=DEFAULT_PROTOCOL,
        )
        main_address = self._process_result_message(
            await self._call(actor_ref.address, control_message)
        )
        real_actor_ref = await self.actor_ref(actor_ref)
        if real_actor_ref.address == main_address:
            raise ValueError("Cannot kill actor on main pool")
        stop_message = ControlMessage(
            new_message_id(),
            real_actor_ref.address,
            ControlMessageType.stop,
            # default timeout (3 secs) and force
            (3.0, force),
            protocol=DEFAULT_PROTOCOL,
        )
        # stop server
        result = await self._call(main_address, stop_message)
        return self._process_result_message(result)

    async def actor_ref(self, *args, **kwargs):
        actor_ref = create_actor_ref(*args, **kwargs)
        local_actor_ref = create_local_actor_ref(actor_ref.address, actor_ref.uid)
        if local_actor_ref is not None:
            return local_actor_ref
        message = ActorRefMessage(
            new_message_id(), actor_ref, protocol=DEFAULT_PROTOCOL
        )
        future = await self._call(actor_ref.address, message, wait=False)
        result = await self._wait(future, actor_ref.address, message)
        return self._process_result_message(result)

    async def send(
        self,
        actor_ref: ActorRef,
        message: Tuple,
        wait_response: bool = True,
        profiling_context: ProfilingContext = None,
    ):
        message = SendMessage(
            new_message_id(),
            actor_ref,
            message,
            protocol=DEFAULT_PROTOCOL,
            profiling_context=profiling_context,
        )

        # use `%.500` to avoid print too long messages
        with debug_async_timeout(
            "actor_call_timeout",
            "Calling %.500r on %s at %s timed out",
            message.content,
            actor_ref.uid,
            actor_ref.address,
        ):
            detect_cycle_send(message, wait_response)
            future = await self._call(actor_ref.address, message, wait=False)
            if wait_response:
                result = await self._wait(future, actor_ref.address, message)
                return self._process_result_message(result)
            else:
                return future

    async def cancel(self, address: str, cancel_message_id: bytes):
        message = CancelMessage(
            new_message_id(), address, cancel_message_id, protocol=DEFAULT_PROTOCOL
        )
        result = await self._call(address, message)
        return self._process_result_message(result)

    async def wait_actor_pool_recovered(self, address: str, main_address: str = None):
        if main_address is None:
            # get main_pool_address
            control_message = ControlMessage(
                new_message_id(),
                address,
                ControlMessageType.get_config,
                "main_pool_address",
                protocol=DEFAULT_PROTOCOL,
            )
            main_address = self._process_result_message(
                await self._call(address, control_message)
            )

        # if address is main pool, it is never recovered
        if address == main_address:
            return

        control_message = ControlMessage(
            new_message_id(),
            address,
            ControlMessageType.wait_pool_recovered,
            None,
            protocol=DEFAULT_PROTOCOL,
        )
        self._process_result_message(await self._call(main_address, control_message))

    async def get_pool_config(self, address: str):
        control_message = ControlMessage(
            new_message_id(),
            address,
            ControlMessageType.get_config,
            None,
            protocol=DEFAULT_PROTOCOL,
        )
        return self._process_result_message(await self._call(address, control_message))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..base.ravel import ravel
from ..datasource.array import asarray
from .concatenate import concatenate


def append(arr, values, axis=None):
    """
    Append values to the end of an array.

    Parameters
    ----------
    arr : array_like
        Values are appended to a copy of this array.
    values : array_like
        These values are appended to a copy of `arr`.  It must be of the
        correct shape (the same shape as `arr`, excluding `axis`).  If
        `axis` is not specified, `values` can be any shape and will be
        flattened before use.
    axis : int, optional
        The axis along which `values` are appended.  If `axis` is not
        given, both `arr` and `values` are flattened before use.

    Returns
    -------
    append : Tensor
        A copy of `arr` with `values` appended to `axis`.  Note that
        `append` does not occur in-place: a new array is allocated and
        filled.  If `axis` is None, `out` is a flattened array.

    See Also
    --------
    insert : Insert elements into an array.
    delete : Delete elements from an array.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.append([1, 2, 3], [[4, 5, 6], [7, 8, 9]]).execute()
    array([1, 2, 3, ..., 7, 8, 9])

    When `axis` is specified, `values` must have the correct shape.

    >>> mt.append([[1, 2, 3], [4, 5, 6]], [[7, 8, 9]], axis=0).execute()
    array([[1, 2, 3],
           [4, 5, 6],
           [7, 8, 9]])
    >>> mt.append([[1, 2, 3], [4, 5, 6]], [7, 8, 9], axis=0)
    Traceback (most recent call last):
        ...
    ValueError: all the input tensors must have same number of dimensions

    """
    arr = asarray(arr)
    if axis is None:
        if arr.ndim != 1:
            arr = arr.ravel()
        values = ravel(values)
        axis = arr.ndim - 1
    return concatenate((arr, values), axis=axis)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialization.serializables import (
    KeyField,
    StringField,
    TupleField,
    DictField,
    BoolField,
)
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_series


class SeriesDatetimeMethod(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATETIME_METHOD

    _input = KeyField("input")
    _method = StringField("method")
    _method_args = TupleField("method_args")
    _method_kwargs = DictField("method_kwargs")
    _is_property = BoolField("is_property")

    def __init__(
        self,
        method=None,
        method_args=None,
        method_kwargs=None,
        is_property=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _method=method,
            _method_args=method_args,
            _method_kwargs=method_kwargs,
            _is_property=is_property,
            _output_types=output_types,
            **kw
        )
        if not self.output_types:
            self.output_types = [OutputType.series]

    @property
    def input(self):
        return self._input

    @property
    def method(self):
        return self._method

    @property
    def method_args(self):
        return self._method_args

    @property
    def method_kwargs(self):
        return self._method_kwargs

    @property
    def is_property(self):
        return self._is_property

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, inp):
        return _datetime_method_to_handlers[self._method].call(self, inp)

    @classmethod
    def tile(cls, op):
        return _datetime_method_to_handlers[op.method].tile(op)

    @classmethod
    def execute(cls, ctx, op):
        return _datetime_method_to_handlers[op.method].execute(ctx, op)


class SeriesDatetimeMethodBaseHandler:
    @classmethod
    def call(cls, op, inp):
        empty_series = build_empty_series(inp.dtype)
        if op.is_property:
            test_obj = getattr(empty_series.dt, op.method)
        else:
            test_obj = getattr(empty_series.dt, op.method)(
                *op.method_args, **op.method_kwargs
            )
        dtype = test_obj.dtype
        return op.new_series(
            [inp],
            shape=inp.shape,
            dtype=dtype,
            index_value=inp.index_value,
            name=inp.name,
        )

    @classmethod
    def tile(cls, op):
        out = op.outputs[0]

        out_chunks = []
        for series_chunk in op.input.chunks:
            chunk_op = op.copy().reset_key()
            out_chunks.append(
                chunk_op.new_chunk(
                    [series_chunk],
                    shape=series_chunk.shape,
                    dtype=out.dtype,
                    index=series_chunk.index,
                    index_value=series_chunk.index_value,
                    name=series_chunk.name,
                )
            )

        params = out.params
        params["chunks"] = out_chunks
        params["nsplits"] = op.input.nsplits
        new_op = op.copy()
        return new_op.new_tileables([op.input], kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        inp = ctx[op.input.key]
        try:
            out = getattr(inp.dt, op.method)
        except ValueError:
            # fail due to buffer read-only
            out = getattr(inp.copy().dt, op.method)
        if not op.is_property:
            out = out(*op.method_args, **op.method_kwargs)
        ctx[op.outputs[0].key] = out


_datetime_method_to_handlers = {}
for method in dir(pd.Series.dt):
    if not method.startswith("_"):
        _datetime_method_to_handlers[method] = SeriesDatetimeMethodBaseHandler



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import itertools
import logging
import time
from collections import namedtuple, OrderedDict
from dataclasses import dataclass
from typing import Dict, Optional, Tuple, Union

from .... import oscar as mo
from .... import resource as mars_resource
from ....typing import BandType
from ...cluster import QuotaInfo

logger = logging.getLogger(__name__)

QuotaDumpType = namedtuple("QuotaDumpType", "allocations requests hold_sizes")


@dataclass
class QuotaRequest:
    req_size: Tuple
    delta: int
    req_time: float
    event: asyncio.Event


class QuotaActor(mo.Actor):
    @classmethod
    def gen_uid(cls, band_name: str):
        return f"{band_name}_quota"

    def __init__(self, band: BandType, quota_size: int, **kw):
        super().__init__()
        self._requests = OrderedDict()

        self._cluster_api = None

        self._band = band
        self._band_name = band[1]

        self._quota_size = quota_size
        self._allocations = dict()
        self._total_allocated = 0

        self._hold_sizes = dict()
        self._total_hold = 0

        if kw:  # pragma: no cover
            logger.warning("Keywords for QuotaActor %r not used", list(kw.keys()))

    async def __post_create__(self):
        from ...cluster.api import ClusterAPI

        try:
            self._cluster_api = await ClusterAPI.create(self.address)
            self._report_quota_info()
        except mo.ActorNotExist:
            pass

    async def _has_space(self, delta: int):
        return self._total_allocated + delta <= self._quota_size

    def _log_allocate(self, msg: str, *args, **kwargs):
        args += (self._total_allocated, self._quota_size)
        logger.debug(msg + " Allocated: %s, Total size: %s", *args, **kwargs)

    def _report_quota_info(self):
        if self._cluster_api is not None:
            quota_info = QuotaInfo(
                quota_size=self._quota_size,
                allocated_size=self._total_allocated,
                hold_size=self._total_hold,
            )
            asyncio.create_task(
                self._cluster_api.set_band_quota_info(self._band_name, quota_info)
            )

    async def request_batch_quota(self, batch: Dict):
        """
        Request for resources in a batch
        :param batch: the request dict in form {request_key: request_size, ...}
        :return: if request is returned immediately, return True, otherwise False
        """
        all_allocated = True
        # check if the request is already allocated
        for key, size in batch.items():
            if key not in self._allocations or size > self._allocations.get(key):
                all_allocated = False
                break

        self._log_allocate("Receive batch quota request %r on %s.", batch, self.uid)
        sorted_req = sorted(batch.items(), key=lambda tp: tp[0])
        keys = tuple(tp[0] for tp in sorted_req)
        quota_sizes = tuple(tp[1] for tp in sorted_req)
        delta = sum(v - self._allocations.get(k, 0) for k, v in batch.items())

        # if all requested and allocation can still be applied, apply directly
        if all_allocated and await self._has_space(delta):
            self._log_allocate("Quota request %r already allocated.", batch)
            return

        if delta > self._quota_size:
            raise ValueError(
                f"Cannot allocate quota size {delta} "
                f"larger than total capacity {self._quota_size}."
            )

        if keys in self._requests:
            event = self._requests[keys].event
        else:
            has_space = await self._has_space(delta)
            if has_space and not self._requests:
                # if no previous requests, we can apply directly
                self._log_allocate(
                    "Quota request met for key %r on %s.", keys, self.uid
                )
                await self.alter_allocations(keys, quota_sizes, allocate=True)
                return
            else:
                # current free space cannot satisfy the request, the request is queued
                if not has_space:
                    self._log_allocate(
                        "Quota request unmet for key %r on %s.", keys, self.uid
                    )
                else:
                    self._log_allocate(
                        "Quota request queued for key %r on %s.", keys, self.uid
                    )
                event = asyncio.Event()
                quota_request = QuotaRequest(quota_sizes, delta, time.time(), event)
                if keys not in self._requests:
                    self._requests[keys] = quota_request

        async def waiter():
            try:
                await event.wait()
            except asyncio.CancelledError as ex:
                await self.ref().remove_requests.tell(keys)
                raise ex

        return waiter()

    async def remove_requests(self, keys: Tuple):
        self._requests.pop(keys, None)
        await self._process_requests()

    def hold_quotas(self, keys: Tuple):
        """
        Mark request quota as already been hold

        Parameters
        ----------
        keys : Tuple
            request keys
        """
        for key in keys:
            try:
                alloc_size = self._allocations[key]
            except KeyError:
                continue
            self._total_hold += alloc_size - self._hold_sizes.get(key, 0)
            self._hold_sizes[key] = alloc_size

    async def release_quotas(self, keys: Tuple):
        """
        Release allocated quota in batch

        Parameters
        ----------
        keys : Tuple
            request keys
        """
        total_alloc_size = 0

        for key in keys:
            try:
                alloc_size = self._allocations.pop(key)
                total_alloc_size += alloc_size
            except KeyError:
                continue
            self._total_hold -= self._hold_sizes.pop(key, 0)

        self._total_allocated -= total_alloc_size
        if total_alloc_size:
            await self._process_requests()

            self._report_quota_info()
            self._log_allocate("Quota keys %s released on %s.", keys, self.uid)

    def dump_data(self):
        return QuotaDumpType(self._allocations, self._requests, self._hold_sizes)

    def get_allocated_size(self):
        # get total allocated size, for debug purpose
        return self._total_allocated

    async def alter_allocations(
        self,
        keys: Tuple,
        quota_sizes: Tuple,
        handle_shrink: bool = True,
        allocate: bool = False,
    ):
        """
        Alter multiple requests

        Parameters
        ----------
        keys : Tuple
            keys to update
        quota_sizes : Tuple
            new quota sizes, if None, no changes will be made
        handle_shrink : bool
            if True and the quota size less than the original, process requests in the queue
        allocate : bool
            if True, will allocate resources for new items
        """
        quota_sizes = quota_sizes or itertools.repeat(None)
        total_old_size, total_diff = 0, 0
        for k, s in zip(keys, quota_sizes):
            old_size = self._allocations.get(k, 0)
            size_diff = 0

            if not allocate and k not in self._allocations:
                total_old_size += old_size
                continue

            if s != old_size:
                s = int(s)
                size_diff = s - old_size
                self._total_allocated += size_diff
                self._allocations[k] = s
                try:
                    self._total_hold += s - self._hold_sizes[k]
                    self._hold_sizes[k] = s
                except KeyError:
                    pass

            total_old_size += old_size
            total_diff += size_diff
        if handle_shrink and total_diff < 0:
            await self._process_requests()

        self._report_quota_info()
        self._log_allocate(
            "Quota keys %r applied on %s. Total old Size: %s, Total diff: %s,",
            keys,
            self.uid,
            total_old_size,
            total_diff,
        )

    async def _process_requests(self):
        """
        Process quota requests in the queue
        """
        removed = []
        for k, req in self._requests.items():
            if await self._has_space(req.delta):
                await self.alter_allocations(
                    k, req.req_size, handle_shrink=False, allocate=True
                )
                req.event.set()
                removed.append(k)
            else:
                # Quota left cannot satisfy the next request, we quit
                break
        for k in removed:
            self._requests.pop(k, None)


class MemQuotaActor(QuotaActor):
    """
    Actor handling worker memory quota
    """

    def __init__(
        self,
        band: BandType,
        quota_size: int,
        hard_limit: int = None,
        refresh_time: Union[int, float] = None,
        enable_kill_slot: bool = True,
    ):
        super().__init__(band, quota_size)
        self._hard_limit = hard_limit
        self._last_memory_available = 0
        self._refresh_time = refresh_time or 1

        self._enable_kill_slot = enable_kill_slot

        self._stat_refresh_task = None
        self._slot_manager_ref = None

    async def __post_create__(self):
        await super().__post_create__()
        self._stat_refresh_task = self.ref().update_mem_stats.tell_delay(
            delay=self._refresh_time
        )

        from .workerslot import BandSlotManagerActor

        try:
            self._slot_manager_ref = await mo.actor_ref(
                uid=BandSlotManagerActor.gen_uid(self._band[1]), address=self.address
            )
        except mo.ActorNotExist:  # pragma: no cover
            pass

    async def __pre_destroy__(self):
        self._stat_refresh_task.cancel()

    async def update_mem_stats(self):
        """
        Refresh memory usage
        """
        cur_mem_available = mars_resource.virtual_memory().available
        if cur_mem_available > self._last_memory_available:
            # memory usage reduced: try reallocate existing requests
            await self._process_requests()
        self._last_memory_available = cur_mem_available
        self._report_quota_info()
        self._stat_refresh_task = self.ref().update_mem_stats.tell_delay(
            delay=self._refresh_time
        )

    async def _has_space(self, delta: int):
        if self._hard_limit is None:
            return await super()._has_space(delta)

        mem_stats = mars_resource.virtual_memory()
        # calc available physical memory
        available_size = (
            mem_stats.available
            - max(0, mem_stats.total - self._hard_limit)
            - (self._total_allocated - self._total_hold)
        )
        if max(delta, 0) >= available_size:
            logger.warning(
                "%s met hard memory limitation: request %d, available %d, hard limit %d",
                self.uid,
                delta,
                available_size,
                self._hard_limit,
            )

            if self._enable_kill_slot and self._slot_manager_ref is not None:
                logger.info("Restarting free slots to obtain more memory")
                await self._slot_manager_ref.restart_free_slots()
            return False
        return await super()._has_space(delta)

    def _log_allocate(self, msg: str, *args, **kwargs):  # pragma: no cover
        if logger.getEffectiveLevel() > logging.DEBUG:
            return

        if self._hard_limit is None:
            return super()._log_allocate(msg, *args, **kwargs)

        mem_stats = mars_resource.virtual_memory()
        # calc available physical memory
        available_size = (
            mem_stats.available
            - max(0, mem_stats.total - self._hard_limit)
            - (self._total_allocated - self._total_hold)
        )
        args += (
            self._total_allocated,
            self._quota_size,
            mem_stats.available,
            available_size,
            self._hard_limit,
            self._total_hold,
        )

        logger.debug(
            msg
            + " Allocated: %s, Quota size: %s, Phy available: %s, Hard available: %s,"
            " Hard limit: %s, Holding: %s",
            *args,
            **kwargs,
        )


class WorkerQuotaManagerActor(mo.Actor):
    def __init__(self, default_config: Dict, band_configs: Optional[Dict] = None):
        self._cluster_api = None
        self._default_config = default_config
        self._band_configs = band_configs or dict()

        self._band_quota_refs = dict()  # type: Dict[str, mo.ActorRef]

    async def __post_create__(self):
        from ...cluster.api import ClusterAPI

        self._cluster_api = await ClusterAPI.create(self.address)

        band_to_resource = await self._cluster_api.get_bands()
        for band in band_to_resource.keys():
            band_config = self._band_configs.get(band[1], self._default_config)
            hard_limit = band_config.get("hard_limit")
            actor_cls = MemQuotaActor if hard_limit else QuotaActor
            self._band_quota_refs[band] = await mo.create_actor(
                actor_cls,
                band,
                **band_config,
                uid=MemQuotaActor.gen_uid(band[1]),
                address=self.address,
            )

    async def __pre_destroy__(self):
        await asyncio.gather(
            *[mo.destroy_actor(ref) for ref in self._band_quota_refs.values()]
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialization.serializables import ListField
from ...tensor.base.sort import _validate_sort_psrs_kinds
from ..utils import parse_index, validate_axis, build_concatenated_rows_frame
from ..core import IndexValue
from .core import DataFrameSortOperand
from .psrs import DataFramePSRSOperandMixin, execute_sort_values


class DataFrameSortValues(DataFrameSortOperand, DataFramePSRSOperandMixin):
    _op_type_ = OperandDef.SORT_VALUES

    by = ListField("by", default=None)

    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)

    @classmethod
    def _tile_dataframe(cls, op):
        df = build_concatenated_rows_frame(op.inputs[0])

        if df.chunk_shape[op.axis] == 1:
            out_chunks = []
            for chunk in df.chunks:
                chunk_op = op.copy().reset_key()
                out_chunks.append(
                    chunk_op.new_chunk(
                        [chunk],
                        shape=chunk.shape,
                        index=chunk.index,
                        index_value=op.outputs[0].index_value,
                        columns_value=chunk.columns_value,
                        dtypes=chunk.dtypes,
                    )
                )
            new_op = op.copy()
            kws = op.outputs[0].params.copy()
            kws["nsplits"] = df.nsplits
            kws["chunks"] = out_chunks
            return new_op.new_dataframes(op.inputs, **kws)
        else:
            if op.na_position != "last":  # pragma: no cover
                raise NotImplementedError("Only support puts NaNs at the end.")
            # use parallel sorting by regular sampling
            return (yield from cls._tile_psrs(op, df))

    @classmethod
    def _tile_series(cls, op):
        series = op.inputs[0]
        if len(series.chunks) == 1:
            chunk = series.chunks[0]
            chunk_op = op.copy().reset_key()
            out_chunks = [
                chunk_op.new_chunk(
                    series.chunks,
                    shape=chunk.shape,
                    index=chunk.index,
                    index_value=op.outputs[0].index_value,
                    dtype=chunk.dtype,
                    name=chunk.name,
                )
            ]
            new_op = op.copy()
            kws = op.outputs[0].params.copy()
            kws["nsplits"] = series.nsplits
            kws["chunks"] = out_chunks
            return new_op.new_seriess(op.inputs, **kws)
        else:
            if op.na_position != "last":  # pragma: no cover
                raise NotImplementedError("Only support puts NaNs at the end.")
            # use parallel sorting by regular sampling
            return (yield from cls._tile_psrs(op, series))

    @classmethod
    def _tile(cls, op):
        inp = op.inputs[0]
        if inp.shape[op.axis] == 0:
            # if the length is zero, return input directly
            return inp
        if inp.ndim == 2:
            return (yield from cls._tile_dataframe(op))
        else:
            return (yield from cls._tile_series(op))

    @classmethod
    def execute(cls, ctx, op: "DataFrameSortValues"):
        in_data = ctx[op.inputs[0].key]
        result = execute_sort_values(in_data, op)
        if op.nrows is not None:
            result = result.head(op.nrows)
        ctx[op.outputs[0].key] = result

    def __call__(self, a):
        assert self.axis == 0
        if self.ignore_index:
            index_value = parse_index(pd.RangeIndex(a.shape[0]))
        else:
            if isinstance(a.index_value.value, IndexValue.RangeIndex):
                index_value = parse_index(pd.Index([], dtype=np.int64))
            else:
                index_value = a.index_value
        if a.ndim == 2:
            return self.new_dataframe(
                [a],
                shape=a.shape,
                dtypes=a.dtypes,
                index_value=index_value,
                columns_value=a.columns_value,
            )
        else:
            return self.new_series(
                [a], shape=a.shape, dtype=a.dtype, index_value=index_value, name=a.name
            )


def dataframe_sort_values(
    df,
    by,
    axis=0,
    ascending=True,
    inplace=False,
    kind="quicksort",
    na_position="last",
    ignore_index=False,
    parallel_kind="PSRS",
    psrs_kinds=None,
):
    """
    Sort by the values along either axis.

    Parameters
    ----------
    df : Mars DataFrame
         Input dataframe.
    by : str
         Name or list of names to sort by.
    axis : %(axes_single_arg)s, default 0
         Axis to be sorted.
    ascending : bool or list of bool, default True
         Sort ascending vs. descending. Specify list for multiple sort
         orders.  If this is a list of bools, must match the length of
         the by.
    inplace : bool, default False
         If True, perform operation in-place.
    kind : {'quicksort', 'mergesort', 'heapsort'}, default 'quicksort'
         Choice of sorting algorithm. See also ndarray.np.sort for more
         information.  `mergesort` is the only stable algorithm. For
         DataFrames, this option is only applied when sorting on a single
         column or label.
    na_position : {'first', 'last'}, default 'last'
         Puts NaNs at the beginning if `first`; `last` puts NaNs at the
         end.
    ignore_index : bool, default False
         If True, the resulting axis will be labeled 0, 1, …, n - 1.
    parallel_kind : {'PSRS'}, default 'PSRS'
         Parallel sorting algorithm, for the details, refer to:
         http://csweb.cs.wfu.edu/bigiron/LittleFE-PSRS/build/html/PSRSalgorithm.html

    Returns
    -------
    sorted_obj : DataFrame or None
        DataFrame with sorted values if inplace=False, None otherwise.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({
    ...     'col1': ['A', 'A', 'B', np.nan, 'D', 'C'],
    ...     'col2': [2, 1, 9, 8, 7, 4],
    ...     'col3': [0, 1, 9, 4, 2, 3],
    ... })
    >>> df.execute()
        col1 col2 col3
    0   A    2    0
    1   A    1    1
    2   B    9    9
    3   NaN  8    4
    4   D    7    2
    5   C    4    3

    Sort by col1

    >>> df.sort_values(by=['col1']).execute()
        col1 col2 col3
    0   A    2    0
    1   A    1    1
    2   B    9    9
    5   C    4    3
    4   D    7    2
    3   NaN  8    4

    Sort by multiple columns

    >>> df.sort_values(by=['col1', 'col2']).execute()
        col1 col2 col3
    1   A    1    1
    0   A    2    0
    2   B    9    9
    5   C    4    3
    4   D    7    2
    3   NaN  8    4

    Sort Descending

    >>> df.sort_values(by='col1', ascending=False).execute()
        col1 col2 col3
    4   D    7    2
    5   C    4    3
    2   B    9    9
    0   A    2    0
    1   A    1    1
    3   NaN  8    4

    Putting NAs first

    >>> df.sort_values(by='col1', ascending=False, na_position='first').execute()
        col1 col2 col3
    3   NaN  8    4
    4   D    7    2
    5   C    4    3
    2   B    9    9
    0   A    2    0
    1   A    1    1
    """

    if na_position not in ["last", "first"]:  # pragma: no cover
        raise TypeError(f"invalid na_position: {na_position}")
    axis = validate_axis(axis, df)
    if axis != 0:
        raise NotImplementedError("Only support sort on axis 0")
    psrs_kinds = _validate_sort_psrs_kinds(psrs_kinds)
    by = by if isinstance(by, (list, tuple)) else [by]
    if isinstance(ascending, list):  # pragma: no cover
        if all(ascending):
            # all are True, convert to True
            ascending = True
        elif not any(ascending):
            # all are False, convert to False
            ascending = False
    op = DataFrameSortValues(
        by=by,
        axis=axis,
        ascending=ascending,
        inplace=inplace,
        kind=kind,
        na_position=na_position,
        ignore_index=ignore_index,
        parallel_kind=parallel_kind,
        psrs_kinds=psrs_kinds,
        gpu=df.op.is_gpu(),
        output_types=[OutputType.dataframe],
    )
    sorted_df = op(df)
    if inplace:
        df.data = sorted_df.data
    else:
        return sorted_df


def series_sort_values(
    series,
    axis=0,
    ascending=True,
    inplace=False,
    kind="quicksort",
    na_position="last",
    ignore_index=False,
    parallel_kind="PSRS",
    psrs_kinds=None,
):
    """
    Sort by the values.

    Sort a Series in ascending or descending order by some
    criterion.

    Parameters
    ----------
    series : input Series.
    axis : {0 or 'index'}, default 0
        Axis to direct sorting. The value 'index' is accepted for
        compatibility with DataFrame.sort_values.
    ascending : bool, default True
        If True, sort values in ascending order, otherwise descending.
    inplace : bool, default False
        If True, perform operation in-place.
    kind : {'quicksort', 'mergesort' or 'heapsort'}, default 'quicksort'
        Choice of sorting algorithm. See also :func:`numpy.sort` for more
        information. 'mergesort' is the only stable  algorithm.
    na_position : {'first' or 'last'}, default 'last'
        Argument 'first' puts NaNs at the beginning, 'last' puts NaNs at
        the end.
    ignore_index : bool, default False
         If True, the resulting axis will be labeled 0, 1, …, n - 1.

    Returns
    -------
    Series
        Series ordered by values.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> raw = pd.Series([np.nan, 1, 3, 10, 5])
    >>> s = md.Series(raw)
    >>> s.execute()
    0     NaN
    1     1.0
    2     3.0
    3     10.0
    4     5.0
    dtype: float64

    Sort values ascending order (default behaviour)

    >>> s.sort_values(ascending=True).execute()
    1     1.0
    2     3.0
    4     5.0
    3    10.0
    0     NaN
    dtype: float64

    Sort values descending order

    >>> s.sort_values(ascending=False).execute()
    3    10.0
    4     5.0
    2     3.0
    1     1.0
    0     NaN
    dtype: float64

    Sort values inplace

    >>> s.sort_values(ascending=False, inplace=True)
    >>> s.execute()
    3    10.0
    4     5.0
    2     3.0
    1     1.0
    0     NaN
    dtype: float64

    Sort values putting NAs first
    """
    if na_position not in ["last", "first"]:  # pragma: no cover
        raise TypeError(f"invalid na_position: {na_position}")
    axis = validate_axis(axis, series)
    if axis != 0:
        raise NotImplementedError("Only support sort on axis 0")
    psrs_kinds = _validate_sort_psrs_kinds(psrs_kinds)
    op = DataFrameSortValues(
        axis=axis,
        ascending=ascending,
        inplace=inplace,
        kind=kind,
        na_position=na_position,
        ignore_index=ignore_index,
        parallel_kind=parallel_kind,
        psrs_kinds=psrs_kinds,
        output_types=[OutputType.series],
        gpu=series.op.is_gpu(),
    )
    sorted_series = op(series)
    if inplace:
        series.data = sorted_series.data
    else:
        return sorted_series



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def copy(a, order="K"):
    """
    Return a tensor copy of the given object.

    Parameters
    ----------
    a : array_like
        Input data.
    order : {'C', 'F', 'A', 'K'}, optional
        Controls the memory layout of the copy. 'C' means C-order,
        'F' means F-order, 'A' means 'F' if `a` is Fortran contiguous,
        'C' otherwise. 'K' means match the layout of `a` as closely
        as possible. (Note that this function and :meth:`ndarray.copy` are very
        similar, but have different default values for their order=
        arguments.)

    Returns
    -------
    arr : Tensor
        Tensor interpretation of `a`.

    Notes
    -----
    This is equivalent to:

    >>> import mars.tensor as mt

    >>> mt.array(a, copy=True)  #doctest: +SKIP

    Examples
    --------
    Create an array x, with a reference y and a copy z:

    >>> x = mt.array([1, 2, 3])
    >>> y = x
    >>> z = mt.copy(x)

    Note that, when we modify x, y changes, but not z:

    >>> x[0] = 10
    >>> (x[0] == y[0]).execute()
    True
    >>> (x[0] == z[0]).execute()
    False

    """
    from ..datasource import array

    return array(a, order=order, copy=True)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def ndim(a):
    """
    Return the number of dimensions of a tensor.

    Parameters
    ----------
    a : array_like
        Input tebsir.  If it is not already a tensor, a conversion is
        attempted.

    Returns
    -------
    number_of_dimensions : int
        The number of dimensions in `a`.  Scalars are zero-dimensional.

    See Also
    --------
    ndarray.ndim : equivalent method
    shape : dimensions of tensor
    Tensor.shape : dimensions of tensor

    Examples
    --------
    >>> import mars.tensor as mt
    >>> mt.ndim([[1,2,3],[4,5,6]])
    2
    >>> mt.ndim(mt.array([[1,2,3],[4,5,6]]))
    2
    >>> mt.ndim(1)
    0

    """
    from ..datasource import asarray

    try:
        return a.ndim
    except AttributeError:
        return asarray(a).ndim



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
from .....resource import Resource
from ..api import ExecutionConfig, register_config_cls
from ..utils import get_band_resources_from_config


@register_config_cls
class MarsExecutionConfig(ExecutionConfig):
    name = "mars"

    def __init__(self, execution_config: Dict):
        super().__init__(execution_config)
        self._mars_execution_config = execution_config[self.backend]

    def get_deploy_band_resources(self) -> List[Dict[str, Resource]]:
        return get_band_resources_from_config(self._mars_execution_config)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import defaultdict
from typing import Dict, List

from ....config import options
from ....core import ChunkGraph
from ....core.operand import VirtualOperand
from ....typing import BandType, ChunkType, OperandType


class Coloring:
    """
    Coloring a chunk graph according to an algorithm
    described in https://github.com/mars-project/mars/issues/2435
    """

    def __init__(
        self,
        chunk_graph: ChunkGraph,
        all_bands: List[BandType],
        chunk_to_bands: Dict[ChunkType, BandType],
        initial_same_color_num: int = None,
        as_broadcaster_successor_num: int = None,
    ):
        self.chunk_graph = chunk_graph
        self.all_bands = all_bands
        self.chunk_to_bands = chunk_to_bands
        if initial_same_color_num is None:
            initial_same_color_num = max(options.combine_size // 2, 1)
        self.initial_same_color_num = initial_same_color_num
        if as_broadcaster_successor_num is None:
            as_broadcaster_successor_num = options.combine_size * 2
        self.successor_same_color_num = as_broadcaster_successor_num

        self._coloring_iter = itertools.count()

    def _next_color(self) -> int:
        return next(self._coloring_iter)

    @classmethod
    def _can_color_same(cls, chunk: ChunkType, predecessors: List[ChunkType]) -> bool:
        if (
            # VirtualOperand cannot be fused
            any(isinstance(n.op, VirtualOperand) for n in [chunk] + predecessors)
            # allocated on different bands
            or len({n.op.gpu for n in [chunk] + predecessors}) > 1
            # expect worker changed
            or len({n.op.expect_worker for n in [chunk] + predecessors}) > 1
            # scheduling hint tells that cannot be fused
            or (
                chunk.op.scheduling_hint is not None
                and not chunk.op.scheduling_hint.can_be_fused()
            )
        ):
            return False
        return True

    def _color_init_nodes(self) -> Dict[OperandType, int]:
        # for initial op with same band but different priority
        # we color them w/ different colors,
        # to prevent from wrong fusion.
        # e.g. md.read_csv ensure incremental index by generating
        # chunks with ascending priorities (smaller one has higher priority),
        # chunk 0 has higher priority than chunk 1,
        # so that when chunk 1 executing, it would know chunk 0's shape
        # TODO: make it general instead handle priority as a special case
        band_priority_to_colors = dict()
        for chunk, band in self.chunk_to_bands.items():
            band_priority = (band, chunk.op.priority)
            if band_priority not in band_priority_to_colors:
                band_priority_to_colors[band_priority] = self._next_color()

        band_priority_to_color_list = defaultdict(list)
        for (band, priority), color in band_priority_to_colors.items():
            band_priority_to_color_list[band, priority].append(color)
        color_to_size = defaultdict(lambda: 0)
        op_to_colors = dict()
        for chunk, band in self.chunk_to_bands.items():
            priority = chunk.op.priority
            color = band_priority_to_color_list[band, priority][-1]
            size = color_to_size[color]
            if size >= self.initial_same_color_num:
                color = self._next_color()
                band_priority_to_color_list[band, priority].append(color)
            color_to_size[color] += 1
            op_to_colors[chunk.op] = color
        return op_to_colors

    def color(self) -> Dict[ChunkType, int]:
        chunk_to_colors = dict()

        # step 1: Coloring the initial nodes according to the bands that assigned by assigner
        op_to_colors = self._color_init_nodes()

        # step2: Propagate color in the topological order,
        # if the input nodes have same color, color it with the same color;
        # otherwise, color with a new color.
        broadcaster_chunk_set = set()
        for chunk in self.chunk_graph.topological_iter():
            if self.chunk_graph.count_successors(chunk) > self.successor_same_color_num:
                # is broadcaster
                broadcaster_chunk_set.add(chunk)

            if chunk.op in op_to_colors:
                # colored
                chunk_to_colors[chunk] = op_to_colors[chunk.op]
                continue

            predecessors = self.chunk_graph.predecessors(chunk)
            pred_colors = {op_to_colors[pred.op] for pred in predecessors}
            if len(predecessors) == 1 and predecessors[0] in broadcaster_chunk_set:
                # TODO: handle situation that chunks which specify reassign_workers
                # predecessor is broadcaster, just allocate a new color
                color = self._next_color()
            elif len(pred_colors) == 1:
                if self._can_color_same(chunk, predecessors):
                    # predecessors have only 1 color, will color with same one
                    color = next(iter(pred_colors))
                else:
                    color = self._next_color()
            else:
                # has more than 1 color, color a new one
                assert len(pred_colors) > 1
                color = self._next_color()

            op_to_colors[chunk.op] = chunk_to_colors[chunk] = color

        # step 3: Propagate with reversed topological order,
        # check a node with its inputs, if all inputs have different color with itself, skip;
        # otherwise, if some of inputs have the same color, but some others have different color,
        # color the input nodes with same one with a new color, and propagate to its inputs and so on.
        for chunk in self.chunk_graph.topological_iter(reverse=True):
            pred_colors = {
                op_to_colors[pred.op]
                for pred in self.chunk_graph.iter_successors(chunk)
            }
            chunk_color = chunk_to_colors[chunk]
            if chunk_color in pred_colors and len(pred_colors) > 1:
                # conflict
                # color the successors with new colors
                stack = []
                for succ in self.chunk_graph.iter_successors(chunk):
                    if chunk_to_colors[succ] == chunk_color:
                        new_color = op_to_colors[succ.op] = self._next_color()
                        for c in succ.op.outputs:
                            if c not in self.chunk_graph:  # pragma: no cover
                                continue
                            chunk_to_colors[c] = new_color
                            stack.extend(self.chunk_graph.successors(c))
                # color the descendants with same color to the new one
                # the descendants will not be visited more than 2 times
                while len(stack) > 0:
                    node = stack.pop()
                    node_color = chunk_to_colors[node]
                    if node_color == chunk_color:
                        # same color, recolor to the new one
                        node_pred_colors = list(
                            {
                                op_to_colors[inp.op]
                                for inp in self.chunk_graph.iter_predecessors(node)
                            }
                        )
                        node_input_same_color = len(node_pred_colors) == 1
                        if node_input_same_color:
                            node_new_color = node_pred_colors[0]
                        else:
                            node_new_color = self._next_color()
                        op_to_colors[node.op] = node_new_color
                        for c in node.op.outputs:
                            if c not in self.chunk_graph:  # pragma: no cover
                                continue
                            chunk_to_colors[c] = node_new_color
                            stack.extend(self.chunk_graph.successors(c))

        return chunk_to_colors
